/* See cpk_grasp.h */
/* Last edited on 2008-01-09 00:10:40 by stolfi */

#define _GNU_SOURCE
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>

#include <cpk_grasp.h>
#include <cpk_mis.h>
#include <cpk_graph.h>
#include <cpk_greedy.h>
#include <cpk_lopt.h>
#include <cpk_valid.h>
#include <cpk_debug.h>
#include <cpk_io.h>
#include <cpk_basic.h>

#include <bool.h>
#include <vec.h>
#include <affirm.h>
#include <jsmath.h>

typedef struct cpk_seed_pair_t 
  { int v1;
    int v2;
    double escore;
  } cpk_seed_pair_t;
  /* Estrutura usada para encontrar pares de vrtices independentes de
    alto peso e baixo consumo. */

/* PROTTIPOS */

i2_vec_t cpk_grasp_get_seed_pairs(cpk_graph_t *G, double W[], int maxPares, bool_t verbose);
  /* Devolve uma lista {K} de pares de vrtices independentes de {G},
     para serem usados como solues de partida da fase de construo
     do GRASP. A lista ter no mximo {maxPares} pares. */

double cpk_grasp_pair_score
  ( cpk_graph_t *G, 
    double W[], 
    int u, 
    int v, 
    bool_t temp[]
  );
  /* Calcula o escore de um par de vrtices {u,v} considerado como
    conjunto inicial para a fase de construo do GRASP. O escore 
    -INF se {u} e {v} so vizinhos, caso contrrio  a soma dos pesos
    de {u} e {v} dividida pelo tamanho da unio de suas vizinhanas
    (incluindo {u} e {v}).
    
    O vetor de booleanos {temp}  usado internamente; ele deve ser
    todo FALSE na entrada e ser todo FALSE na sada. */

/* IMPLEMENTAES */

#define VALIDATE TRUE

int_vec_t cpk_grasp_find_indep_set
  ( cpk_graph_t *G,
    double W[], 
    double maxClock,  /* Tempo limite para retornar a resposta. */
    int seed,         /* Semente para escolhas aleatrias. */
    bool_t verbose
  )
  {
    /* As fases de construo e busca sao repetidas para varios pares
    de vrtices independentes de partida, ciclicamente. */

    if (verbose) { cpk_trace_entry(); }

    int nV = G->nV;

    /* Melhor soluo encontrada at o momento (inicialmente vazia): */
    int *SM = (int *)notnull(malloc(nV*sizeof(int)), "no mem");
    int nSM = 0;     /* A melhor soluo  {SM[0..nSM-1]}. */
    double WSM = 0;  /* Peso total de {SM}. */

    /* Inicializa gerador de nmeros aleatrios: */
    srand(seed);
    
    if (nV > 0)
      { /* Decide quantos pares-semente usar. Melhor ficar com poucos e bons, acho: */
        int maxSeedPairs = (int)ceil(sqrt((double)nV));
        /* Constri um conjunto de no mximo {maxSeedPairs} pares de vrtices iniciais: */
        i2_vec_t seedPair = cpk_grasp_get_seed_pairs(G, W, maxSeedPairs, verbose);
        if (verbose) { cpk_trace(); }
        
        /* Representao do conjunto independente {S} e admissvel {Q}: */
        cpk_mis_state_t *SQ = cpk_mis_state_new(G, W);

        /* Roda vrias tentativas GRASP: */
        int nIter = 0;
        do
          { if (verbose)  { fprintf(stderr, "Iterao %d\n", nIter); }

            /* FASE DE CONSTRUO */

            /* Esvazia {S}, coloca todos os ns em {Q}, ordenados com base em WR: */
            cpk_mis_init(SQ, verbose);

            if (nIter == 0)
              { /* Na iterao 0, comea com a pilha vazia: */
              }
            else
              { /* Incializa {S} com uma semente variada: */
                if (verbose) { fprintf(stderr, "\n  ------------------------\n"); }
                if (seedPair.ne > 0)
                  { /* Inicializa {S} com o prximo par-semente: */
                    if (verbose) { fprintf(stderr, "  Inserindo um par inicial"); }
                    { int t = nIter % seedPair.ne;
                      int u = seedPair.e[t].c[0];
                      int v = seedPair.e[t].c[1];
                      cpk_mis_add_indep(SQ, u, verbose);
                      cpk_mis_add_indep(SQ, v, verbose);
                    }
                  }
                else 
                  { /* Inicializa {S} com o prximo vrtice: */
                    int u = nIter % nV;
                    cpk_mis_add_indep(SQ, u, verbose);
                  }
                if (verbose) { fprintf(stderr, "\n"); }
                if (verbose) { TRACE_SOL("  Soluo semente", SQ->nS, SQ->WS);  }
              }
            /* Expande a soluo aleatrio-gulosamente at ficar maximal: */
            if (verbose) { fprintf(stderr, "\n  ------------------------\n"); }
            if (verbose) { fprintf(stderr, "  Expanso gulosa randomizada:"); }
            cpk_greedy_maximize_random(SQ, verbose);
            if (verbose) { fprintf(stderr, "\n");  }
            if (verbose) { TRACE_SOL("  Soluo gulosa randomizada", SQ->nS, SQ->WS);  }

            if (VALIDATE) { cpk_check_indep_set(G, SQ->nS, SQ->S, NULL); }

            /* Atualiza a melhor soluo: */
            cpk_lopt_update_best_solution(SQ, &nSM, SM, &WSM, verbose);

            /* FASE DE BUSCA */

            /* At quantos vrtices vale a pena tentar relocar por vez? */
            int maxRem = 3;

            /* Faz otimizao local com retirada de mltiplos vrtices: */
            assert(pqueue_count(SQ->Q) == 0);
            if (verbose) { fprintf(stderr, "\n  ------------------------\n"); }
            if (verbose) { fprintf(stderr, "  Otimizao local:"); }
            cpk_local_opt(SQ, maxRem,  &nSM, SM, &WSM, verbose);
            if (verbose) { fprintf(stderr, "\n");  }
            if (verbose) { TRACE_SOL("  Soluo localmente otimizada", SQ->nS, SQ->WS);  }

            if (VALIDATE)  
              { cpk_check_indep_set(G, SQ->nS, SQ->S, NULL);
                cpk_check_indep_set(G, nSM, SM, NULL);
              }

            /* Mais uma iterao: */
            nIter++;
          }
        while(cpk_cpu_time_1() < maxClock);

        /* Copia melhor soluo para soluo corrente: */
        cpk_mis_copy_indep(SQ, nSM, SM, verbose);
        assert(cpk_weight_cmp(SQ->WS, WSM) == 0);
        if (verbose) { fprintf(stderr, "\n  ------------------------\n"); }
        if (verbose) { TRACE_SOL("  Soluo tima bruta", SQ->nS, SQ->WS);  }

        /* Faz busca local com max 1 remoo, para garantir: */
        if (verbose) { fprintf(stderr, "\n  ------------------------\n"); }
        if (verbose) { fprintf(stderr, "  Re-otimizao local:");  }
        cpk_local_opt(SQ, 1, &nSM, SM, &WSM, verbose);
        if (verbose) { fprintf(stderr, "\n");  }

        /* Copia melhor soluo para soluo corrente: */
        cpk_mis_copy_indep(SQ, nSM, SM, verbose);
        assert(cpk_weight_cmp(SQ->WS, WSM) == 0);
        if (verbose) { TRACE_SOL("  Soluo tima otimizada", SQ->nS, SQ->WS);  }

        /* Limpeza: */
        cpk_mis_state_free(SQ);
      }

    if (VALIDATE) { cpk_check_indep_set(G, nSM, SM, NULL); }

    if (verbose)
      { TRACE_SOL("Soluo GRASP final", nSM, WSM);
        cpk_print_vertex_set(stderr, "", nSM, SM, NULL, 30);
      }

    /* prepara a saida da rotina */
    int_vec_t J = int_vec_new(nSM); 
    { int i; for (i = 0; i < nSM; i++) { J.e[i] = SM[i]; } }

    if (verbose) { cpk_trace_exit(); }
    return J;
  }

i2_vec_t cpk_grasp_get_seed_pairs(cpk_graph_t *G, double W[], int maxPares, bool_t verbose) 
  {
    if (verbose) { cpk_trace_entry(); }
    int nV = G->nV;
    int totPares = (nV-1)*nV/2;
    if (verbose) { fprintf(stderr, "  Grafo com %d vrtices (%d pares)\n", nV, totPares); }

    /* Obtm lista de vrtices {X[0..nV-1]}} em ordem decrescente de escore: */
    int X[nV];
    { int v; for(v = 0; v < nV; v++) { X[v] = v; } }
    
    /* Ordena lista por escore decrescente: */
    auto int compara_vertices(const void *a, const void *b);
    qsort(X, nV, sizeof(int), &compara_vertices);

    int compara_vertices(const void *a, const void *b)
      { int u = *((int*)a); double eu = W[u]/(G->deg[u]+1);
        int v = *((int*)b); double ev = W[v]/(G->deg[v]+1);
        return -dblcmp(eu,ev); /* Como so {doubles} no podemos retornar {eu-ev}: */
      }

    /* Quantos pares vale a pena olhar no mximo? (n^2  demais!) */
    int maxParesExm = imin(10*nV, totPares);
    if (verbose) { fprintf(stderr, "  Coletando %d pares-semente\n", maxParesExm); }
    cpk_seed_pair_t *uvs;
    uvs = (cpk_seed_pair_t *)notnull(malloc(maxParesExm*sizeof(cpk_seed_pair_t)), "no mem");
    int nPares = 0;
    { bool_t temp[nV]; /* Temporrio para clculo dos escores. */
      { int v; for (v = 0; v < nV; v++) { temp[v] = FALSE; } }
      int s, j;
      /* Enumera pares : */
      for(s = 0; (s < nV) && (nPares < maxParesExm); s++)
        { for(j = 0; (j <= s) && (nPares < maxParesExm); j++)
            { int u = X[j];
              int v = X[s - j];
              double escore = cpk_grasp_pair_score(G, W, u, v, temp);
              if (escore > -INF)
                { uvs[nPares] = (cpk_seed_pair_t){ u, v, escore};
                  nPares++;
                }
            }
        }
      if (verbose) { fprintf(stderr, "  Encontrou %d pares-semente\n", nPares); }
       
      /* Ordena o vetor {uvs[0..nPares-1]} em ordem decrescente de escore: */
      auto int compara_escores(const void *a, const void *b);
      qsort(uvs, nPares, sizeof(cpk_seed_pair_t), &compara_escores);

      int compara_escores(const void *a, const void *b)
        { double ea = ((cpk_seed_pair_t *)a)->escore;
          double eb = ((cpk_seed_pair_t *)b)->escore;
          /* Como escores so {double}s, no podemos simplesmente subtrair: */
          if (ea > eb)
            { return -1; }
          else if (ea < eb)
            { return +1; }
          else
            { return 0; }
        }
    }

    /* Devolve os melhores {maxPares} pares: */
    nPares = imin(nPares, maxPares);
    i2_vec_t par = i2_vec_new(nPares);
    { int i;
      for (i = 0; i < nPares; i++)
        { cpk_seed_pair_t *pi = &(uvs[i]);
          par.e[i] = (i2_t){{ pi->v1, pi->v2 }};
        }
    }
    
    /* Limpeza: */
    free(uvs);
    
    if (verbose) { cpk_trace_exit(); }
    return par;
  }

double cpk_grasp_pair_score
  ( cpk_graph_t *G, 
    double W[], 
    int u, 
    int v, 
    bool_t temp[]
  )
  {
    /* Parania: */
    if (u == v) { return -INF; }
    
    /* Pega graus e vizinhanas de {u} e {v}: */
    int du = G->deg[u]; int *nbu = &(G->nbr[G->fnb[u]]);
    int dv = G->deg[v]; int *nbv = &(G->nbr[G->fnb[v]]);
    
    /* Supe que {temp} est toda {FALSE} incialmente. */
    /* Liga {temp[w]} sse {w}  vizinho de {u}: */
    { int j;  for (j = 0; j < du; j++) { assert(! temp[nbu[j]]); temp[nbu[j]] = TRUE; } }
    
    /* Calcula escore: */
    double escore;
    if (temp[v])
      { /* {v}  vizinho de {u}. */
        escore = -INF;
      }
    else
      { /* Conta vizinhos de {v} que so vizinhos de {u}: */
        int duv = 0;
        { int j; for (j = 0; j < dv; j++) { if (temp[nbv[j]]) { duv++; } } }
        if (duv > 0)
          { /* Calcula tamanho {tamViz} da unio das vizinhanas: */
            int tamViz = du + dv + 2 - duv; /* Comea contando {u}, {v}, e os vizinhos de {u}. */
            /* Escore  razo peso/consumo: */
            escore = (W[u]+W[v]) / ((double)tamViz);
          }
        else
          { /* Distncia de {u} a {v}  maior que 2: */
            escore = -INF;
          }
      }
      
    /* Limpa {temp} para futuras chamadas: */
    { int j; for (j = 0; j < du; j++) { temp[nbu[j]] = FALSE; } }
    
    return escore; 
  }
