/* See hxg_canvas.h */
/* Last edited on 2009-02-10 09:18:46 by stolfi */ 

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include <bool.h>
#include <i2.h>
#include <r2.h>
#include <interval.h>

#include <hxg_canvas.h>
#include <cpk_graph.h>
#include <cpk_basic.h>

r2_t hxg_pixel_pos(hxg_canvas_t *cvs, i2_t p)
  {
    int jx = X(p), iy = Y(p); 
    double sx = X(cvs->step), sy = Y(cvs->step);
    double x = X(cvs->org) + (jx + (double)(iy & 1)/2.0)*sx;
    double y = Y(cvs->org) + iy*sy;
    return (r2_t){{x,y}};
  }
  
void hxg_get_proximity_graph
  ( hxg_canvas_t *cvs, /* Canvas. */
    bool_t valid[],      /* Which pixels to take. */
    double pixW[],     /* Weight of each pixel. */
    r2_vec_t *V,       /* OUT: Vertex coordinates. */
    double_vec_t *W,   /* OUT: Value of each vertex. */
    double dLim,       /* Distance threshold. */
    i2_vec_t *E        /* OUT: Proximity edges. */
  )
  {
    int nx = X(cvs->size), ny = Y(cvs->size); 
    int nPix = nx*ny; int k; 
    /* Count nonzero pixels, build index {I}: */
    int nV = 0;
    int_vec_t I = int_vec_new(nPix); /* Linear pixel index {k} to vertex number {iV}. */
    { for (k = 0; k < nPix; k++) 
        { if (valid[k]) { I.e[k] = nV; nV++; } else { I.e[k] = -1; } }
    }
    /* fprintf(stderr, "Counted %d nonzero pixels.\n", nV); */

    /* Gather their indices and positions, in sweep order (Y, then X): */
    /* Also replace the pixel value by {i+1} where {i} is the index in {V}. */
    i2_vec_t K = i2_vec_new(nV); /* Vertex number {iV} to pixel indices {ix,iy}. */
    (*V) = r2_vec_new(nV);
    (*W) = double_vec_new(nV);
    { int iy;
      for (iy = 0; iy < ny; iy++)
        { int k = iy*nx;
          int ix;
          for (ix = 0; ix < nx; ix++,k++)
            { if (valid[k]) 
                { i2_t u = (i2_t){{ix,iy}};
                  int iV = I.e[k];
                  K.e[iV] = u;
                  V->e[iV] = hxg_pixel_pos(cvs, u);
                  W->e[iV] = pixW[k];
                }
            }
        }
      /* fprintf(stderr, "Collected %d nonzero pixels.\n", nV); */
    }
    
    /* Get neighborhood templates {D[0],D[1]} for even and odd scanlines: */
    i2_vec_t D[2];
    { int r; 
      for (r = 0; r < 2; r++)
        { D[r] = hxg_disk_arcs(cvs, (i2_t){{0,r}}, dLim);
          /* fprintf(stderr, "Template for (iy%%2)=%d has %d arcs.\n", r, D[r].ne); */
        }
    }
    /* Get the proximity edges. */
    { /* Allocate the edge vector: */
      (*E) = i2_vec_new(V->ne*D[0].ne); /* Estimated size. */
      int nE = 0;
      int iV;
      for (iV = 0; iV < nV; iV++)
        { /* Get pixel indices {ix,iy} of {V[iV]}: */
          i2_t *Ki = &(K.e[iV]);
          int ix = X(*Ki), iy = Y(*Ki); 
          /* Get template edges appropriate for scanline {iy}: */
          i2_vec_t *Di = &(D[iy & 1]);
          int nD = Di->ne;
          /* Apply template to {ix,iy}: */
          int s;
          for (s = 0; s < nD; s++)
            { /* Get edge {(ex,ey)} from template: */
              i2_t *e = &(Di->e[s]);
              int ex = ORG(*e), ey = DST(*e); 
              /* Compute destination pixel {jx,jy}: */
              int jx = ix + ex, jy = iy + ey;
              if ((jx >= 0) && (jx < nx) && (jy >= 0) && (jy < ny))
                { /* Get the pixel's linearized index {k}: */
                  int k = jy*nx + jx;
                  if (valid[k])
                    { /* Destination is valid, get its index {j} in {V}: */
                      int jV = I.e[k];
                      /* Sanity checks: */
                      assert(jV < nV);
                      assert(jV > iV);
                      assert(X(K.e[jV]) == jx);
                      assert(Y(K.e[jV]) == jy);
                      /* Add edge to {E}: */
                      i2_vec_expand(E, nE); /* Just to be sure... */
                      E->e[nE] = (i2_t){{iV,jV}};
                      nE++;
                    }
                }
            }
        }
      i2_vec_trim(E, nE);
    }
    
    /* Housecleaning: */
    free(D[0].e);
    free(D[1].e);
    free(K.e);
    free(I.e);
  }

i2_vec_t hxg_disk_arcs(hxg_canvas_t *cvs, i2_t c, double r)
  {
    /* Get hold of params: */
    double sx = X(cvs->step), sy = Y(cvs->step);  /* Steps of canvas's grid. */
    int ciy = Y(c);
    double rad2 = r*r;
    /* Compute largest displacement {eymax} in Y: */  
    int eymax = (int)floor(r/sy);
    /* Conservative estimate of template size: */
    int mD = ((int)floor((3.15/2)*r/sx)+1)*(eymax+1);
    /* Allocate the template and fill it: */
    i2_vec_t D = i2_vec_new(mD);
    int nD = 0;
    int ey;
    for (ey = 0; ey <= eymax; ey++)
      { int iy = ciy + ey;
        double dy = ey*sy;
        /* Compute circle half-extent {hx} on this scanline: */
        double dy2 = dy*dy;
        double hx = (dy2 > rad2 ? 0.0 : sqrt(rad2 - dy2));
        /* Get index range {exmin..exmax} of pixels inside interval: */
        double fx = ((iy & 1) - (ciy & 1))/2.0; /* Rel. scanline shift */
        int exmin = (int)floor(-hx/sx - fx)+1;
        int exmax = (int)ceil(hx/sx - fx)-1;
        /* On scanline {ciy}, consider only forward pixels: */
        if (ey == 0) { exmin = 1; }
        /* Enumerate interior pixels on this scanline: */
        int ex;
        for (ex = exmin; ex <= exmax; ex++)
          { i2_t e = (i2_t){{ex,ey}};
            i2_vec_expand(&D, nD); /* Just in case... */
            D.e[nD] = e; nD++;
          }
      }
    i2_vec_trim(&D, nD);
    return D;
  }

double hxg_sum_over_circle
  ( hxg_canvas_t *cvs, 
    r2_t c, 
    double r, 
    double f(int ix, int iy, r2_t *p),
    r2_t *p
  )
  { 
    /* Get hold of params: */
    double ox = X(cvs->org), oy = Y(cvs->org);    /* Lower left corner of grid. */
    double sx = X(cvs->step), sy = Y(cvs->step);  /* Steps of canvas's grid. */
    double cx = X(c), cy = Y(c);
    double r2 = r*r;
    /* Compute range {iymin..iymax} of scanlines for {dMAx} circle at {ox,oy}. */
    int iymin = (int)ceil((cy - r - oy)/sy);
    int iymax = (int)ceil((cy + r - oy)/sy)-1;
    /* Enumerate relevant scanlines: */
    int iy;
    double fsum = 0;
    for (iy = iymin; iy <= iymax; iy++)
      { /* Displacement of scanline {iy} from center. */
        double dy = oy + iy*sy - cy;
        /* Find half-extent {hx} of circle on scanline {iy}: */
        double dy2 = dy*dy;
        double hx = (dy2 > r2 ? 0.0 : sqrt(r2 - dy2));
        /* Get range {ixmin..ixmax} of circle pixels on this scanline: */
        /* We assume that the circle is infinitesimally shifted to the left: */
        int ixmin = (int)ceil((cx - hx - ox)/sx - (iy & 1)/2.0);
        int ixmax = (int)ceil((cx + hx - ox)/sx - (iy & 1)/2.0) - 1;
        int ix; 
        for (ix = ixmin; ix <= ixmax; ix++) 
          { fsum += f(ix,iy,p); }
      }
    return fsum;
  }
