#ifndef hxg_canvas_H
#define hxg_canvas_H

/* Pixel arrays with hexagonal geometry. */
/* Last edited on 2008-01-09 00:27:53 by stolfi */

#define _GNU_SOURCE
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>

#include <bool.h>
#include <i2.h>
#include <r2.h>
#include <interval.h>

#define MAX_CANVAS_SIZE 2048
/* Maximum width or height of an hexagonal canvas. */

typedef struct hxg_canvas_t 
  { r2_t org;         /* Position of lower left grid point. */
    r2_t step;        /* Grid steps. */
    i2_t size;        /* Number of grid points along each axis. */    
  } hxg_canvas_t;
  /* A {hxg_canvas_t} is an array of "pixels", associated with points
    of an hexagonal grid. The bottom left point of the grid is {org};
    points are spaced {step[0]} along each horizontal row, and each
    row is spaced {step[1]} vertically from the previous one. Both
    steps are positive. Odd rows are displaced by {0.5*step[0]} in the
    positive direction. Each row has {size[0]} points, and there are
    {size[1]} rows.
    
    Storage area for the pixels should be provided separately by
    clients. The standard way is to provide a vector {pix[0..N-1]}
    where {N = X(size)*Y(size)}, where the pixels are stored by
    scanlines -- i.e. pixel {ix,iy} is stored at {pix[ix +
    X(size)*iy]}. */
    
r2_t hxg_pixel_pos(hxg_canvas_t *cvs, i2_t p);
  /* Returns the nominal coordinates of the point number {p.c[0]} on
    row {p.c[1]} of the grid of canvas {cvs}. */

void hxg_get_proximity_graph
  ( hxg_canvas_t *cvs, /* Canvas. */
    bool_t valid[],      /* Which pixels to take. */
    double pixW[],     /* Weight of each pixel. */
    r2_vec_t *V,       /* OUT: Vertex coordinates. */
    double_vec_t *W,   /* OUT: Value of each vertex. */
    double dLim,       /* Distance threshold. */
    i2_vec_t *E        /* OUT: Proximity edges. */
  );
  /* Takes a canvas {cvs} and two pixels vectors {valid} and {pixW},
    stored in the standard way. Returns an undirected graph {(V,E)}
    whose vertices are the positions of all pixels {k} of {cvs} with
    {valid[k] = TRUE}, with an edge {(i,j)} whenever {i<j} and
    {dist(V[i],V[j]) < dLim}. Also sets {W[i]} to the pixel's value
    {pixW[k]}. The vectors {V}, {W}, and {E} are allocated by the
    procedure. */

/* 
  ENUMERATION OF THE INFINITE GRID
  
  The following procedures consider the infinite hex grid underlying 
  the canvas {cvs}, ignoring its bounds. */

i2_vec_t hxg_disk_arcs(hxg_canvas_t *cvs, i2_t c, double r);      
  /* Returns a list of integer vectors {D[k]} such that the pixel with
    indices {p + D[k]} lies at distance less than {r} from the pixel
    with indices {c}. Lists all pairs {D[k] = (ex,ey)} such that {ey >
    0}, or {ey = 0} and {ex > 0}. Note that the set {D} varies
    depending on whether {c} lies on an  even or odd scanline. */

double hxg_sum_over_circle
  ( hxg_canvas_t *cvs, 
    r2_t c, 
    double r, 
    double f(int ix, int iy, r2_t *p),
    r2_t *p
  );
  /* Computes the sum of {f(ix,iy,p)} over all grid pixels {ix,iy} that lie
    inside the circle with center at the pixel {c} and radius {r}. */

#endif
