
/*
  
  */
  
typedef int32_t dg_wing_index_t;
  /* Identifies a cell slot in the star of some locus {E}. */

#define dg_NOT_WING ((dg_wing_index_t)-1)
  /* The {dg_wing_index_t} of a cell that is not a wing. */

bool_t dg_cell_intersects_star(dg_dim_t d, dg_cell_index_t k, dg_locus_t E);
  /* Returns TRUE if cell {k} intersects some cell of {K(E)}. */
   
void dg_star_box_root_relative(dg_dim_t d, dg_locus_t E, interval_t B[]);
  /* Stores in {B[0..d-1]} the lower and upper root-relative
    coordinates of the star of locus {E} in a {d}-dimensional grid.
    
    The upper endpoints of will be in the range {(0_1)}. Due to the
    toroidal geometry, the lower endpoint {LO(B[i])} will be negative
    whenever {cell(E)} has position 0 along axis {i}. In particular,
    if {rank(E)<d}, then {B[i]} will be set to {[-1,+1]} for every
    axis {i} in {rank(E)..d-1} */
  
/*
  VALID LOCUS
  
  A locus {E} of a dyadic grid {G} is /valid/ if all the cells
  of its star {K(E)} are cells of {G}. Note that each item of {G}
  may be associated to zero or more valid loci.
  
  MINIMAL VALID LOCUS
  
  A locus {E} of a dyadic grid {G} is /minimal valid/ if it is valid
  and there is no other valid locus {E'} of {G}, with the same item
  as {E}, whose star {K(E')} is smaller than {K(E)} (i.e., whose rank 
  is greater than that of {E}). It can be seen that {E} is minimal valid if and only if
  it is valid, and at least one of the cells in {K(E)} is a leaf of
  {G}.
  
  MAXIMAL VALID LOCUS
  
  Similarly, a locus {E} of a dyadic grid {G} is /maximal valid/ if it
  is valid and there is no other valid locus {E'} of {G}, with the
  same item as {E}, whose star {K(E')} is greater than {K(E)} (i.e.,
  whose rank is less than that of {E}). It can be seen that {E} is
  maximal valid if and only if it is valid, and its rank is that of
  the first layer where that item appears.
  
  ENUMERATION OF MINIMAL AND MAXIMAL LOCI
  
  In what follows, {E} is a locus in some given dyadic greid {G},
  {NE} is the node star {N(E,T)} of {E} in the given grid, and 
  {dtE} is an arbitrary client data record associated with {E}. */

typedef void dg_locus_data_t;
  /* An arbitrary client data record associated with a locus. */
  
typedef void dg_locus_visit_proc_t
  ( dg_locus_t E, 
    dg_rank_t r, 
    dg_node_star_t *NE,
    dg_locus_data_t *dtE, 
    bool_t maxE
  );
  /* A procedure used by {dg_enum_critical_locus} to report a locus {E}
    of rank {r} that is maximal valid (if {maxE=TRUE}) or minimal
    valid (if {maxE=FALSE}). */

typedef dg_locus_data_t *dg_root_locus_data_proc_t(dg_locus_t E, dg_node_t *root);
  /* A procedure used by {dg_enum_critical_locus} to create an
    initial client data record for a locus {E} of rank 0. */
    
typedef dg_locus_data_t *dg_shrink_locus_data_proc_t
  ( dg_locus_t E, 
    dg_rank_t r, 
    dg_axis_t splax,
    dg_node_star_t *NE,
    dg_locus_data_t *dtE
  );
  /* A procedure used by {dg_enum_critical_locus} to create a new
    client data record for the child {E'} of a non-leaf valid locus
    {E} of rank {r} whose item {B} also exists in level {r+1}. Each
    each cell of the star of {E} gets split in perpendicular to axis
    {splax}. */

typedef dg_locus_data_t *dg_split_locus_data_proc_t
  ( dg_locus_t E, 
    dg_rank_t r, 
    dg_axis_t splax,
    dg_node_star_t *NE,
    dg_locus_data_t *dtE,
    dg_locus_data_t **dtLO, 
    dg_locus_data_t **dtHI, 
    dg_locus_data_t **dtMD
  );
  /* A procedure used by {dg_enum_critical_locus} to create new client
    data records {dtLO,dtHI,dtMD} for the children {ELO,EHI,EMD} of a
    non-leaf valid locus {E} of rank {r} which gets split in level
    {r+1}. Children {ELO} and {EHI} are the low and high halves of
    {E}, and {EMD} is their shared facet. The cut is perpendicular to
    axis {splax}. */
    
typedef void dg_free_locus_data_proc_t
  ( dg_locus_t E, 
    dg_rank_t r, 
    dg_node_star_t *NE, 
    dg_locus_data_t *dtE
  );
  /* A procedure used by {dg_enum_critical_locus} to reclaim the
    space used by the client data record {dtE}, created by 
    one of the calls above. */

void dg_enum_critical_loci
  ( dg_dim_t d, 
    dg_node_t *root,
    dg_locus_visit_proc_t *visit, 
    dg_root_locus_data_proc_t *root_data,
    dg_shrink_locus_data_proc_t *shrink_data,
    dg_split_locus_data_proc_t *split_data,
    dg_free_locus_data_proc_t *free_data
  );
  /* Enumerates all maximal and minimal loci of the finite {d}-dimensional
    multigrid {G} represented by the tree rooted at {t}. 
    
    The procedure calls {visit(E, r, dtE, TRUE)} is for each maximal
    valid locus, and {visit(E, r, dtE, FALSE)} for each minimal valid
    locus. The argument {r} given to {visit} is the rank of {E}.
    
    The argument {dtE} is some arbitrary client data record associated
    with {E}. Such records are created by {root_data}, {shrink_data},
    and {split_data} and eventually recovered by {free_data}. */
  
dg_locus_vec_t dg_find_critical_loci(dg_dim_t d, dg_node_t *t, bool_t vtxOnly, bool_t maximal);
  /* Returns a list of all maximal (if {maximal=TRUE}) or minimal
    ({maximal=FALSE}) valid loci of the finite {d}-dimensional
    multigrid {G} represented by the tree rooted at {t}. If {vtxOnly},
    returns only vertex loci. */

/* 
  INDEXING IN A TREE STAR
  
  Let {T(E)} be the tree star of an {m}-dimensional locus {E}, that
  is, an array containing the enclosing nodes of the {2^{d-m}} cells
  in level {r} of {G*} which have {E} as a face. Those cells are
  listed in such an order that cells {T(E)[u]} and {T(E)[v]} are
  adjacent across a hyperplane {H} orthogonal to axis {Nrm(E)[j]},
  with {T(E)[u]} on the low side of {H}, if and only if {u < v} and {u
  \XOR v = 2^j}. ??? Hopefully this is the same as saying that 
  the address of a tree {t} in {T(E)} is the the column-order 
  address of each computed from the position vector
  of {t}'s cell relative to the star of {E}. */
