/* See dg_plotextra.h */
/* Last edited on 2005-02-14 23:12:17 by stolfi */

#include <dg_plot.h>
#include <dg_plotextra.h>
#include <dg_tree.h>
#include <bz_basic.h>
#include <dgfaceshape.h>
#include <affirm.h>
#include <vec.h>
#include <ia.h>
#include <math.h>

/* IMPLEMENTATIONS */

void dg_enum_flattened
  ( dg_dim_t d,            /* Dimension of grid. */
    dg_FaceShape *sh,    /* Shape of face {F}. */
    dg_axis_t ax[],        /* Axes which are perpendicular and parallel to {F}. */
    dg_rank_t r,           /* Rank of some cell of which {F} is face. */
    dg_Index_vec_t K,    /* Cells of rank {r} that are incident to {F}. */
    dg_node_ref_vec_t N,  /* Nodes containing those cells. */
    double tol,          /* Flattening tolerance. */
    dg_rank_t rMax,        /* Never split below this rank. */
    dg_FaceProc process  /* Client face-processing procedure. */  
  )
  { dg_dim_t m = sh->b.m;        /* Dimension of domain of face {F}. */
    dg_dim_t s = d-m;            /* Co-dimension of domain of face {F}. */
    int nFaces = ipow(3, m);   /* Number of faces of {F}. */
    int nCells = ipow(2, s);   /* Number of adjacent cells in {*k,*n}. */
    bool_t split;

    
    fprintf(stderr, "+ dg_enum_flattened(dim = %d r = %d)\n", sh->b.m, r);
    
    /* Check whether splitting-and-recursion is needed. */ 
    /* The face must be split if it is not flat, or if any adjacent cell
      has a split along an axis parallel to the face: */
    if (m == 0)
      { /* A vertex never needs splitting. */
        split = FALSE;
      }
    else if (r >= rMax)
      { /* Too deep to split; force the entire face to be flat: */
        dg_FaceShape_flatten(sh);
        split = FALSE;
      }
    else
      { int i;
        split = (sh->dev.e[0] > 0.0); 
        /* Could be improved --- should skip levels with parallel splits. */
        for (i = 0; (!split) && (i < nCells); i++)
          { split |= ((N.e[i] != NULL) && (N.e[i]->ch[BLO] != NULL)); }
      }

    if (! split) 
      { /* No need to split, process it: */
        dg_Bezier a = dg_Bezier_uniform_new(sh->b.m, sh->b.n, 1);
        dg_Bezier_multiaffine_approx(&(sh->b), &a);
        process(d, &a, ax, r, K, N);
        free(a.c);
      }
    else 
      { /* Split the adjacent cells, recurse: */
        dg_axis_t splax = r % d;  /* Split axis. */
        int splj;               /* Position of split axis in {ax} list. */

        /* Find position {splj} of {splax} in the axis sequence: */
        for (splj = 0; (splj < d) && (ax[splj] != splax); splj++) { }
        affirm(splj < d, "inconsistent ax[]");
        fprintf(stderr, "split on axis ax[%d] = %d\n", splj, splax);

        if (splj < s)
          { /* The split axis is perpendicular to the face. */
            int smask = ipow(2, splj);
            dg_Index_vec_t Kch = dg_Index_vec_new(nCells);
            dg_node_ref_vec_t Nch = dg_node_ref_vec_new(nCells);
            int i;
            /* Split each adj cell, discard halves not incident to face: */
            for (i = 0; i < nCells; i++)
              { int Ki = K.e[i]; dg_node_t *Ni = N.e[i];
                /* Decide which child of {K[i]} is to be kept: */
                int keep = ((i & smask) == 0 ? 1 : 0);
                /* Keep the child of {K[i]} that is adjacent to {F}: */
                if (Ki == dg_NULL_INDEX) 
                  { Kch.e[i] = dg_NULL_INDEX; }
                else
                  { Kch.e[i] = 2*Ki + keep; }
                /* If the node {N[i]} is level with {K[i]}, split it too: */
                if ((Ni == NULL) || (Ki == dg_NULL_INDEX))
                  { Nch.e[i] = NULL; }
                else if (Ni->index < Ki)
                  { Nch.e[i] = Ni; }
                else
                  { /* If {N[i]} is a leaf, keep it: */
                    dg_node_t *Nchi = Ni->ch[keep];
                    Nch.e[i] = (Nchi == NULL ? Ni : Nchi);
                  }
              }
            /* Recurse on same face, with same shape. */
            dg_enum_flattened(d, sh, ax, r+1, Kch, Nch, tol, rMax, process);
            free(Kch.e); free(Nch.e);
          }
        else
          { /* The split axis is parallel to the face. */
            dg_FaceShape shMD, shLO, shHI;
            /* Split face and each adj cell: */
            /* Data for subfaces: */
            ?? dg_Index_vec_t Kch = dg_Index_vec_new(2*nCells);
            ?? dg_node_ref_vec_t Nch = dg_node_ref_vec_new(2*nCells);
            ?? /* Data for splitting facet: */
            ?? dg_FaceShape shMD = dg_FaceShape_facet_new(m, sh->b.n, sh->b.g, splj-s);
            ?? dg_Index_vec_t KMD = (dg_Index_vec_t){ 2*nCells, &(Kch.e[0]) };
            ?? dg_node_ref_vec_t NMD = (dg_node_ref_vec_t){ 2*nCells, &(Nch.e[0]) };
            ?? /* Data for low half-face: */
            ?? dg_FaceShape shLO = dg_FaceShape_new(m, sh->b.n, sh->b.g);
            ?? dg_Index_vec_t KLO = (dg_Index_vec_t){ nCells, &(Kch.e[0]) };
            dg_node_ref_vec_t NLO = (dg_node_ref_vec_t){ nCells, &(Nch.e[0]) };
            ?? /* Data for high half-face: */
            ?? dg_FaceShape shHI = dg_FaceShape_new(m, sh->b.n, sh->b.g);
            ?? dg_Index_vec_t KHI = (dg_Index_vec_t){ nCells, &(Kch.e[nCells]) };
            ?? dg_node_ref_vec_t NHI = (dg_node_ref_vec_t){ nCells, &(Nch.e[nCells]) };
              
            int i;

            /* Split the face shape: */
            dg_FaceShape_split(sh, splj - s, tol, &shLO, &shHI, &shMD);

            /* Recurse for each half of face {F}: */
            dg_enum_flattened(&shLO, tol, rMax, process);
            dg_enum_flattened(&shHI, tol, rMax, process);

            /* Recurse for the separating face {S} of {F}. */
            dg_enum_flattened(&shMD, tol, rMax, process);

            dg_FaceShape_free(* shMD);
            dg_FaceShape_free(* shLO);
            dg_FaceShape_free(* shHI);
            free(Kch.e); 
            free(Nch.e);
          }
      }

    fprintf(stderr, "- dg_enum_flattened(dim = %d r = %d)\n", sh->b.m, r);
    
  }

