#ifndef mdg_pack_H
#define mdg_pack_H

/* Cell blocks (packs) in multi-dimensional dyadic grids */
/* Last edited on 2009-08-23 12:10:44 by stolfi */

#define _GNU_SOURCE
#include <stdlib.h>
#include <stdint.h>

#include <bool.h>
#include <sign.h>
#include <interval.h>
#include <box.h>
#include <vec.h>

#include <udg_grid.h>
#include <mdg_grid.h>

/*
  CELL PACKS
  
  A /cell pack/ is an array {\X} of cells in a given layer {r} of
  {G*}, defined by a position vector {pos[0..d-1]} in that layer and a /pack size
  vector/ {psz[0..d-1]} of positive integers. In the procedures below, a
  pack is usually specified by its size vector {psz} and the cell index
  {b} of the cell whose position is {pos}.
  
  Let {gsz[0..d-1]} be the size of layer {r} along each axis. A cell {C'}
  of layer {r} belongs to the pack {\X} if and only if its position
  vector {pos'} satisfies {pos'[i] = pos[i] + e[i] (mod gsz[i])} for
  every {i}, where each {e[i]} is some integer in the range {0..psz[i]-1}.
  The vector {e} is the /relative position/ of {C'} within the pack.

  DUPLICATE ENTRIES
  
  A cell pack is defined as an array, so it has exactly {N =
  psz[0]*psz[d-1]} entries. This is true even when {psz[i] > gsz[i]} for
  some {i}. In this case the pack wraps around and overlap itself,
  and includes some cells two or more times, with different relative
  position vectors {e}.
  
  PACK INDICES
  
  Each entry of the cell pack is identified by a /slot index/ {pkx}
  in {0..N-1}, which is related to its relative position vector
  {e[0..d-1]} and the pack sizes {psz[0..d-1]} by the function
  {ix_pos_by_col}.
  
  The lower cell of the pack has relative position {e = (0,..0)} and
  slot index {0}. The cell with relative position {e[i] = psz[i]-1} for
  all {i} is the /upper cell/ of the pack, and has slot index {N-1}. */
  
typedef uint64_t mdg_pack_slot_index_t;
  /* Identifies a cell slot in the pack of some locus {E}. */

#define mdg_NOT_SLOT ((mdg_pack_slot_index_t)-1)
  /* The {mdg_pack_slot_index_t} of a cell that is not in the pack. */

uint64_t mdg_pack_slot_count(mdg_dim_t d, mdg_grid_size_t psz[]);
  /* Returns the number of slots in a cell pack with sinze
    {psz[i]} along each axis {i}, namely  {PROD { psz[i] : i \in 0..d-1 }}. */

void mdg_pack_cells
  ( mdg_dim_t d, 
    mdg_cell_index_t b,
    mdg_grid_size_t psz[], 
    mdg_cell_index_t k[]
  );
/* Returns in {k[0..N-1]} the indices of all cells in the pack whose
  lower cell is {b} and whose size along axis {i} is {psz[i]}, for {i}
  in {0..d-1}; where {N = mdg_pack_slot_count(d,psz)}. */

mdg_pack_slot_index_t mdg_pack_find_cell
  ( mdg_dim_t d, 
    mdg_cell_index_t k, 
    mdg_cell_index_t b,
    mdg_grid_size_t psz[]
  );
  /* Returns the slot index {pkx} of cell {k} in the cell pack with
    lower cell {b} and size vector {psz[0..d-1]}. If {k} occurs
    multiple times in the pack, returns the first occurrence. If {k}
    is not part of the pack, returns {mdg_NOT_SLOT} */
   
bool_t mdg_pack_intersects_cell
  ( mdg_dim_t d, 
    mdg_cell_index_t k, 
    mdg_cell_index_t b,
    mdg_grid_size_t psz[]
  );
  /* Returns TRUE if cell {k} intersects some cell of 
    the pack with lower cell {b} and size {psz[0..d-1]}. */
   
void mdg_pack_box_root_relative
  ( mdg_dim_t d, 
    mdg_cell_index_t b,
    mdg_grid_size_t psz[], 
    interval_t B[]
  );
  /* Stores in {B[0..d-1]} the lower and upper coordinates of the box
    of {R^d} that corresponds to the cell pack with lower cell {b}
    and size vector {psz[0..d-1]}.
    
    The lower endpoints will be in the range {[0_1)}; the upper
    endpoint will be strictly greater. Due to the toroidal geometry,
    the upper endpoint {HI(B[i])} will be 1 or more whenever the
    position of {b} along axis {i} is {gsz[i]-psz[i]} or more,
    where {gsz[i]} is the grid's size along that axis. */

/* 
   MISCELLANEOUS */

sign_t mdg_rank_cmp(mdg_rank_t *x, mdg_rank_t *y);
  /* Returns {-1} if {*x < *y}, {0} if {*x == *y}, {+1} if {*x > *y}.
    Useful for sorting things by rank. */ 

/* 
  PRINTOUT */

void mdg_cell_print(FILE *wr, mdg_dim_t d, mdg_cell_index_t k);
  /* Prints the cell {k} in the format 
    "{RANK}:({POS[0]},..{POS[d-1]})" */

#endif


