/* Cell subsequences (packs) in unidimensional dyadic grids. */
/* Last edited on 2009-08-23 11:47:01 by stolfi */

#ifndef udg_pack_H
#define udg_pack_H

#define _GNU_SOURCE
#include <stdint.h>
#include <stdio.h>

#include <interval.h>
#include <bool.h>
#include <sign.h>
#include <vec.h>

#include <udg_grid.h>

/*
  CELL PACKS
  
  A /cell pack/ is a sequence {\X} of consecutive cells in a given
  layer {r} of {G*}, defined by the layer index {r}, an /initial
  position/ {pos}, the position of the lowest cell in that layer, and
  a non-negative /pack size/ {psz}, the number of cells in the pack.
  The pack may fold around the end of the layer.
  Alternatively (and in several procedures below), a pack is specified
  by its size {psz} and by the cell index {b} of the inferior cell.
  
  Let {gsz} be the size of layer {r}. A cell {C'} of layer {r} belongs
  to the pack {\X} if and only if its position {pos'} satisfies {pos'
  = pos + e (mod gsz)}, where {e} is some integer in the range
  {0..psz-1}. 
  
  The integer {e} is the /relative position/ or /slot index/ of {C'}
  within the pack. The lowest (inferior) cell of the pack has slot
  index 0, and the uppermost (superior) has slot index {psz-1}. */
  
typedef int64_t udg_pack_slot_index_t;
  /* Identifies a cell slot in a cell pack.  Normally between 0 and {psz-1}
    where {psz} is the pack size. */

#define udg_NOT_SLOT ((udg_pack_slot_index_t)-1)
  /* The {udg_pack_slot_index_t} of a cell that is not in the pack. */

void udg_pack_cells(udg_cell_index_t b, udg_grid_size_t psz, udg_cell_index_t k[]);
  /* Returns in {k[0..psz-1]} the indices of all cells in the pack 
    whose lower cell is {b} and whose size is {psz}. */

udg_pack_slot_index_t udg_pack_find_cell
  ( udg_cell_index_t k, 
    udg_cell_index_t b,
    udg_grid_size_t psz
  );
  /* Returns the slot index {pkx} of cell {k} in the cell pack with
    lower cell {b} and size {psz}. If {k} occurs
    multiple times in the pack, returns the first occurrence. If {k}
    is not part of the pack, returns {udg_NOT_SLOT} */
   
bool_t udg_pack_intersects_cell
  ( udg_cell_index_t k, 
    udg_cell_index_t b,
    udg_grid_size_t psz
  );
  /* Returns TRUE if cell {k} intersects some cell of 
    the pack with lower cell {b} and size {psz}. */
   
void udg_pack_interval_root_relative
  ( udg_cell_index_t b,
    udg_grid_size_t psz, 
    interval_t *B
  );
  /* Stores in {B} the interval that is the union of all cells in the
    cell pack with lower cell {b} and size {psz}.
    
    The lower endpoint will be in the range {[0_1)}; the upper
    endpoint will be strictly greater. Due to the toroidal geometry,
    the upper endpoint {HI((*B))} will be 1 or more whenever the
    position of {b} is {gsz-psz} or more, where {gsz} is the grid's
    size. */

#endif
