/* General-purpose geometric models. */
/* Last edited on 2005-08-22 23:40:16 by stolfi */

#ifndef gmo_H
#define gmo_H

#include <quad.h>
#include <bool.h>
#include <frgb.h>
#include <hr3.h>

typedef enum 
  { gmo_item_kind_DOT = 0,
    gmo_item_kind_BAR = 1,
    gmo_item_kind_TRI = 2
  } gmo_item_kind_t;
  
typedef int gmo_knot_ix_t;
typedef int gmo_point_ix_t;
typedef int gmo_plane_ix_t;
typedef int gmo_color_ix_t;
typedef int gmo_opto_ix_t;
typedef int gmo_item_ix_t;
  /* Indices into {knot}, {point}, {plane}, {color}, and {opto} tables 
    of a {gmo_t} model. */
  
typedef struct gmo_knot_t
  { gmo_point_ix_t pt;   /* Position (index into {point} list) or -1 if none. */
    gmo_point_ix_t tp;   /* Tangent plane (index into {plane} list) or -1 if none. */
    gmo_opto_ix_t opto;  /* Optical operator (index into {opto} list). */
    float radius;        /* Radius/thickness. */
  } gmo_knot_t; 
  /* A `knot' is an infinitesimal surface element, that is,
    a point of space, a tangent plane, and an opto. */

typedef struct gmo_item_dot_t 
  { gmo_knot_ix_t kn;
  } gmo_item_dot_t; 
  /* A `dot' element, that is, a small sphere. The knot specifies
    the center, radius, and opto of the sphere.
    The tangent plane of the knot is irrelevant. */

typedef struct gmo_item_bar_t 
  { gmo_knot_ix_t kn[2]; 
  } gmo_item_bar_t;
  /* A `bar' item, that is, a sphere-capped cone. The knots specify
    the centers, radii, and optos of the two end-spheres.
    The tangent planes of the knots are irrelevant. */

typedef struct gmo_item_tri_t 
  { gmo_knot_ix_t kn[3]; /* Positions, optos, radii and tangent planes of corners. */
  } gmo_item_tri_t;
  /* A `tri' element, that is, a triangle with rounded edges and corners.
    The knots specify the centers, radii, and optos of the three end-spheres.
    The tangent planes of the knots are irrelevant. */

typedef void *gmo_item_addr_t;
  /* Pointer to a plottable item; actually to a {gmo_dot_t},
    {gmo_bar_t}, or {gmo_tri_t}. */

typedef enum 
  { gmo_opto_kind_EMIT     = 0,
    gmo_opto_kind_ABSORB   = 1,
    gmo_opto_kind_REFLECT  = 2,
    gmo_opto_kind_TRANSMIT = 3
  } gmo_opto_kind_t;
  
typedef struct gmo_opto_t
  {
    frgb_t color;          /* Fraction of coverage per band. */
    gmo_opto_kind_t kind;  /* Kind of optical effect. */
    float sigma;           /* Additional parameter, e.g. roughness. */
    int next;              /* Underlying opto, or -1 if none. */
  } gmo_opto_t;
/* An `opto' (short for `optical operator') specifies how a surface 
  interacts with light at a point. It is assumed to be a linear combination
  of standard bi-directional radiation functions (BDRFs):

    * An EMIT function emits wite light (intensity 1) on all bands.
      If {sigma=0} the light is emitted perpendicularly to the surface;
      if {sigma=1} it is emitted like the Lambertian distribution
      (so that the apparent color is independent of viewing angle).

    * A REFLECT function reflects or diffuses light coming 
      from the same side as the observer. If {sigma=0} the
      surface is a smooth mirror; if {sigma=1} it is an opaque
      Lambertian diffuser.

    * A TRANSMIT function transmits or scatters light coming
      from the side of the surface opposite to the observer.
      If {sigma=0} the surface is clear glass, if {sigma=1}
      it is a translucent Lambertian diffuser. 

    * An ABSORB function emits nothing.
  
  An {opto} is actually a sequence of zero or more `stages', each
  described by an {gmo_opto_t} record. For each band {k}, the
  parameter {color.c[k]} specifies the percentage of the incident view
  rays that hit this stage of the opto. The remaining rays are assumed
  to interact with the {next} stage and following stages. An ABSORB
  stage with {color.c[k]=1} (i.e. a coal black layer) is assumed at
  the end of every stack, to catch any rays that are not caught by
  preceding stages.
  
  Note that rays that hit a TRANSMIT stage go through the surface 
  unchanged, while those that miss it may interact with its 
  subsequent stages. This behavior is not intuitive but 
  simplifies the math.

  Intermediate values of {sigma} give a BDRF that is intermediate between
  the cases {sigma=0} and {sigma=1}. Thus, for example, highlights
  are obtained by adding a REFLECT stage with small but nonzero {sigma}. */

typedef struct gmo_t 
   { int npoints;             /* Number of points. */
     hr3_point_t *point;      /* {point[0..npoints-1]} are homog coords of key points. */
     int nplanes;             /* Number of planes. */
     hr3_plane_t *plane;      /* {plane[0..nplanes-1]} are homog coeffs of key planes. */
     int noptos;              /* Number of optos. */
     gmo_opto_t *opto;        /* Opto list. */
     int nknots;              /* Number of knots. */
     gmo_knot_t *knot;        /* Knot list. */
     int nitems;              /* Number of plottable items. */
     gmo_item_kind_t *iknd;   /* {kind[0..nitems-1]} are the item types. */
     gmo_item_addr_t *item;   /* {item[0..nitems-1]} are the addresses of the items. */
  } gmo_t;

void gmo_write(FILE *wr, gmo_t *P);
  /* Writes a description of {P} {wr}, in a format that can be
    read back with {gmo_read} below.. */
  
gmo_t *gmo_read(FILE *rd);
  /* Reads a plot list {P} from the reader {rd}. Assumes the
    format used by {Write} above. */

/* ITEM ENUMERATION */

typedef bool_t gmo_item_dot_proc_t(gmo_item_ix_t ix, gmo_item_dot_t *t);
typedef bool_t gmo_item_bar_proc_t(gmo_item_ix_t ix, gmo_item_bar_t *t);
typedef bool_t gmo_item_tri_proc_t(gmo_item_ix_t ix, gmo_item_tri_t *t);
  /* Client procedures for {gmo_enum_items}. */

gmo_item_ix_t gmo_enum_items
  ( gmo_t *P, 
    gmo_item_dot_proc_t visit_dot,
    gmo_item_bar_proc_t visit_bar,
    gmo_item_tri_proc_t visit_tri
  );
  /* Enumerates all items of {P} in order, calling the appropriate {visit_XXX}
    procedure on each. If any of the procedures returns {TRUE}, stops
    the enumeration and returns that element's index. Otherwise
    returns -1. */

/* INPUT/OUTPUT */

gmo_item_kind_t gmo_item_kind_from_char(char c);
char gmo_item_kind_to_char(gmo_item_kind_t iknd);

gmo_opto_kind_t gmo_opto_kind_from_char(char c);
char gmo_opto_kind_to_char(gmo_opto_kind_t oknd);

#endif
