/* See {jsjpeg_image.h} */
/* Last edited on 2011-06-15 01:13:13 by stolfi */

/* Created by R. Minetto (IC-UNICAMP) as {ipng.c} sometime in 2008--2009. */
/* Adapted by J. Stolfi (IC-UNICMP) on 2011-05-14. */

#define _GNU_SOURCE
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include <png.h>

#include <bool.h>
#include <affirm.h>
#include <jsfile.h>
#include <jspnm.h>
#include <jspnm_image.h>
#include <jspng_image.h>

pnm_image_t *jspng_image_read (char *name, bool_t verbose, double *gammaP)
  {
    FILE *rd = open_read (name, verbose);
    pnm_image_t *im = jspng_image_fread(rd, gammaP);
    if (rd != stdin) { fclose(rd); }
    return im;
  }

#define PNG_MAGIC_BYTES 8

pnm_image_t *jspng_image_fread (FILE *rd, double *gamma)
  {
    /* Check the PNG signature: */
    unsigned char sig[PNG_MAGIC_BYTES];
    fread(sig, 1, PNG_MAGIC_BYTES, rd);
    demand(png_cmp_sig(sig, 0, PNG_MAGIC_BYTES) == 0, "not a PNG file");

    ???; /* Rewrite handling of sBit, bits. */

    /* Allocate the reader structures: */
    png_structp pr = notnull(png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL), "no mem");
    png_infop pi = notnull(png_create_info_struct(pr), "no mem");

    /* Attach the file handle to the PNG reader, tell it we already read the sig: */
    png_init_io(pr, rd);
    png_set_sig_bytes(pr, PNG_MAGIC_BYTES);

    /* Read the header: */
    png_read_info(pr, pi);

    /* Get attributes: */
    int cols = png_get_image_width(pr, pi);
    int rows = png_get_image_height(pr, pi);
    int chns = png_get_channels(pr, pi);
    int bits = png_get_bit_depth(pr, pi);

    /* Expand 1,2,4-bit images to 1 pixel per byte: */
    if (bits < 8) { png_set_packing(pr); }

    /* Set input transformations: */
    if (color_type == PNG_COLOR_TYPE_PALETTE) { png_set_palette_to_rgb(pr); }
    if ((color_type == PNG_COLOR_TYPE_GRAY) && (bits < 8)) { png_set_gray_1_2_4_to_8(pr); }
    
    /* If palette with transparent color, convert to RGBA: */
    if (png_get_valid(pr, pi, PNG_INFO_tRNS)) { png_set_tRNS_to_alpha(pr); }
    
    /* Get the gamma exponent of the file (but do not do gamma correction): */
    double  gamma;
    if (! png_get_gAMA(pr, pi, gammaP)) { (*gammaP) = NAN; }

    /* Let the library do interlace handling: */
    int number_of_passes = png_set_interlace_handling(pr);

    png_read_update_info(pr, pi);

    /* Allocate scanlines: */
    int png_row_bytes = png_get_row_bytes(pr, pi);
    png_bytep *png_arrayP = notnull(malloc(rows*sizeof(png_bytep)), "no mem");
    int row;
    for (row = 0; row < rows; row++) { png_arrayP[row] = notnull(malloc(png_row_bytes), "no mem"); }

    /* Allocate the PNG image: */
    pnm_image_t *img = pnm_image_new(chns, cols, rows);
    img->maxval = (bits <= 8 ? 255 : 65535);

    /* Samples per row in PNG and PNM images: */
    int spr = cols*chns; /* Samples per row. */

    /* Read whole image: */
    png_read_image(pr, png_arrayP);

    for (row = 0; row < rows; row++)
      { 
        png_bytep png_P = png_arrayP[row];
        pnm_sample_t *pnm_P = img->smp[row];
        int k;
        for (k = 0; k < spr; k++)
          { int smp = (*png_P); png_P++;
            if (bits > 8)
              { /* PNG file has two bytes per sample: */
                smp = (smp << 8) | (*png_P); png_P++;
              }
            (*pnm_P) = smp; pnm_P++;
          }
      }    
    for (row = 0; row < rows; row++) { free(png_arrayP[row]); png_arrayP[row] = NULL; }
    free(png_arrayP);
    png_destroy_read_struct(&pr, &pi, NULL);

    return img; 
  }
  
void jspng_image_fwrite (FILE *wr, pnm_image_t *img, double gamma)
  {

    ???; /* omit gama if neg/zero/nan */
    ???; /* Rewrite logic to chose sBit, bits, and omaxval. */

    int chns = img->chns;
    int cols = img->cols;
    int cols = img->rows;
    
    pnm_sample_t imaxval = img->maxval;  /* Image's {maxval}. */
    demand(imaxval < 65535, "maxval is too big");
    demand(imaxval > 0, "maxval is zero");

    /* Create {libpng} work records {pr, pi}. */
    /* Avoid {setjmp/longjmp} complexity for now, use default error handling. */
    png_struct *pr = notnull(png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL), "no mem");
    png_info *pi = notnull(png_create_info_struct(pr), "no mem");

    /* Assign {wr} as the output file for writes to {pr}: */
    png_init_io(pr, wr);

    /* Select the output bit depth: */
    int bits = (imaxval <= 255 ? 8 : 16);
    
    /* Select the scale factor so that {maxval} always goes to {2^bits-1}. */
    int omaxval = (1<<bits) - 1;
    double scale = ((double)omaxval)/((double)imaxval);
    if (verbose) 
      { if (imaxval != omaxval) { fprintf(stderr, "rescaling from [0..%d] to [0..%d]\n", imaxval, omaxval); }
        if (scale != floor(scale)) { fprintf(stderr, "there may be roundoff errors\n"); }
      }
    
    /* Select output image type: */
    int ctype;
    if (chns == 1)
      { ctype = PNG_COLOR_TYPE_GRAY; }
    else else if (chns == 2)
      { ctype = PNG_COLOR_TYPE_GRAY_ALPHA; }
    else if (chns == 3)
      { ctype = PNG_COLOR_TYPE_RGB; }
    else if (chns == 4)
      { ctype = PNG_COLOR_TYPE_RGB_ALPHA; }
    else
      { demand("invalid number of channels"); }

    png_set_IHDR(pr, pi, cols, rows, bits, ctype, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

    /* Set the {zlib} compression level: */
    png_set_compression_level(pr, 5);

    /* Provide a "gAMA" chunk. */
    png_set_gAMA(pr, pi, gamma);

    /* Write the PNG header: */
    png_write_info (pr, pi);
    
    /* Samples per row in PNG and PNM images: */
    int spr = cols*chns; /* Samples per row. */

    /* Allocate output scanline buffer: */
    png_byte *png_rowP = (png_byte *)malloc(spr*sizeof(png_byte));
    
    /* Process rows: */
    /* Assumes non-interlaced PNG: */
    int row;
    for (row = 0; row < rows; ++row)
      {
        png_bytep png_P = png_rowP;
        pnm_sample_t *pnm_P = img->smp[row];
        int k;
        for (k = 0; k < spr; k++)
          { int smp = (*pnm_P); pnm_P++;
            /* Apply scale factor: */
            if (imaxval != omaxval) { smp = (int)floor(scale*smp+0.5); }
            /* Store in PNG row buffer: */
            if (bits <= 8)
              { /* PNG file has 1 byte per sample: */
                (*png_P) = smp; png_P++;
              }
            else
              { /* PNG file has two bytes per sample: */
                (*png_P) = (smp >> 8); png_P++;
                (*png_P) = (smp & 255); png_P++;
              }
          }
        png_write_row(pr, png_rowP);
      }    
    png_write_end(pr, pi);
    fflush(ofp);
    png_destroy_write_struct(&pr, &pi);
    free(png_rowP);
  }
