/* See pnm_leo.h */
/* Last edited on 2008-05-25 01:41:10 by stolfi */

#include <pnm_leo.h>

/****************
 *
 * Make a new image of a certain size.
 */

Image *NewImage(int width, int height)
{
  Image *image;

  image = (Image *) malloc(sizeof(Image));

  image->width = width;
  image->height = height;
  image->type = Unknown;

  image->data = (u_char *) malloc(width * height);

  return image;
}
  
/****************
 *
 * Read a generic PNM header, returning the type, width, height,
 * and depth of the image.
 */

static int ReadPNMHeader(FILE *fp, int *width, int *height, int *depth)
{
  char ch;
  int type;

  /* determine the file format */

  fscanf(fp, "P%c\n", &ch);

  if (ch >= '1' && ch <= '6') {
    type = (ch <= '3') ? Ascii : Raw;

    if (ch == '1' || ch == '4')
      type |= PBM;
    else if (ch == '2' || ch == '5')
      type |= PGM;
    else
      return Unknown;
  }

  else
    return Unknown;

  /* skip comments */

  do {
    while ((ch = getc(fp)) == '\n');
    if (ch == '#')
      fscanf(fp, "%*[^\n]\n");
  } while (ch == '#');
  ungetc(ch, fp);

  /* read the width, height, and depth if present*/

  if (type & PBM) {
    fscanf(fp, "%d %d\n", width, height);
    *depth = 1;
  }

  else
    fscanf(fp, "%d %d %d\n", width, height, depth);

  return type;
}

/****************************************************************
 *
 * The following functions read the data from a certain type of
 * file into a pre-allocated buffer.
 *
 ****************************************************************/

/****************
 *
 * Read the data from an ASCII PBM file.
 */

static void ReadAsciiPBM(FILE *fp, u_char *data, int width, int height)
{
  int val;
  u_char *cp, *ce;

  cp = data;

  for (ce = cp + width * height; cp < ce; cp++) {
    fscanf(fp, "%d", &val);
    *cp = val ? 255 : 0;
  }
}

/****************
 *
 * Read the data from a raw PBM file.
 */

static void ReadRawPBM(FILE *fp, u_char *data, int width, int height)
{
  u_char *cp, temp;
  int i, j, k;

  cp = data;

  for (j = 0; j < height; j++) {
    for (i = 0; i < width / 8; i++) {
      fread(&temp, 1, 1, fp);
      for (k = 0; k < 8; k++)
        *(cp++) = (temp & (0x80 >> k)) ? 0 : 255;
    }
    if (width % 8 > 0) {
      fread(&temp, 1, 1, fp);
      for (k = 0; k < width % 8; k++)
        *(cp++) = (temp & (0x80 >> k)) ? 0 : 255;
    }
  }
}

/****************
 *
 * Read the data from an ASCII PGM file.
 */

static void ReadAsciiPGM(FILE *fp, u_char *data, int width, int height)
{
  int val;
  u_char *cp, *ce;

  cp = data;

  for (ce = cp + width * height; cp < ce; cp++) {
    fscanf(fp, "%d", &val);
    *cp = val;
  }
}

/****************
 *
 * Read the data from a raw PGM file.
 */

static void ReadRawPGM(FILE *fp, u_char *data, int width, int height)
{
  fread(data, width * height, 1, fp);
}

/****************************************************************
 *
 * Top-level functions to read PNM files into newly-allocated
 * buffers.
 *
 ****************************************************************/

/****************
 *
 * Read any PNM file. Returns a pointer to the new image buffer.
 * Sets the type, width, height, and depth parameters.
 */

Image *ReadPNM(char *fn)
{
  Image *image;
  FILE *fp;
  int width, height, depth, type;

  fp = fn ? fopen(fn, "r") : stdin;
  if (!fp)
    return NULL;

  if ((type = ReadPNMHeader(fp, &width, &height, &depth)) == Unknown)
    return NULL;

  image = (Image *) malloc(sizeof(Image));

  image->type = type;
  image->width = width;
  image->height = height;

  /* fprintf(stderr, "readpnm: type %x\n", type); */

  if (type & PBM) {
    image->data = (u_char *) malloc(width * height);
    if (type & Ascii)
      ReadAsciiPBM(fp, image->data, width, height);
    else
      ReadRawPBM(fp, image->data, width, height);
  }

  else if (type & PGM) {
    image->data = (u_char *) malloc(width * height);
    if (type & Ascii)
      ReadAsciiPGM(fp, image->data, width, height);
    else
      ReadRawPGM(fp, image->data, width, height);

  }

  return image;
}

/****************************************************************
 *
 * Functions to write PNM files. Right now only writes raw
 * files, since I can't see a reason for ASCII files.
 *
 ****************************************************************/

/****************
 *
 * Write a PBM file.
 */

void WritePBM(char *fn, Image *image)
{
  FILE *fp;
  u_char *cp, temp;
  int i, j, k;

  fp = fn ? fopen(fn, "w") : stdout;
  if (!fp)
    return;

  fprintf(fp, "P4\n%d %d\n", image->width, image->height);

  cp = image->data;

  for (j = 0; j < image->height; j++) {
    for (i = 0; i < image->width / 8; i++) {
      for (k = temp = 0; k < 8; k++)
	temp |= (*(cp++) ? (0x80 >> k) : 0);
      fwrite(&temp, 1, 1, fp);
    }
    if (image->width % 8 > 0) {
      for (k = temp = 0; k < image->width % 8; k++)
	temp |= (*(cp++) ? (0x80 >> k) : 0);
      fwrite(&temp, 1, 1, fp);
    }
  }

  fclose(fp);
}

/****************
 *
 * Write a PGM file.
 */

void WritePGM(char *fn, Image *image)
{
  FILE *fp;

  fp = fn ? fopen(fn, "w") : stdout;
  if (!fp)
    return;

  fprintf(fp, "P5\n%d %d %d\n", image->width, image->height, 255);
  fwrite(image->data, image->width * image->height, 1, fp);
  fclose(fp);
}
