/* wavep_leo.c */
/* Last edited on 2008-05-25 02:29:40 by stolfi */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <wavep_leo.h>

#define ISCALE    255.0
#define OSCALE    255.0
#define INTOSCALE 255

int   sizes[MAXL] = {1,2,4,8,16,32,64,128,256,512};
int   levelTable[MAXD];
float scaleTable[MAXL][MAXL];
int   DecompMode = MODE_SPACIAL;

typedef struct Coeff
{
  int index;
  float val;
} Coeff;

typedef int (*Intfunc)();

void GrayInit(void);
void BuildScaleTable(float scale, int minscale);
void FailExit(char *msg);
void CopyRow(float *before, float *after);
void DecompSubRow(float *before, float *after, int size);
void ReconSubRow(float *before, float *after, int size);
void DecompMixedSubRow(float *before, float *after, int size, int p);
void ReconMixedSubRow(float *before, float *after, int size, int p);
void DumpRow(float *data);
void DecompRow(float *data, int toplevel);
void DecompMixedRow(float *data, int toplevel, int roworcol);
void ReconRow(float *data, int toplevel);
void ReconMixedRow(float *data, int toplevel, int roworcol);
void DecompRows(Gray *gray);
void DecompMixedRows(Gray *gray, int roworcol);
void ReconRows(Gray *gray);
void ReconMixedRows(Gray *gray, int roworcol);
void Transpose(Gray *gray);
void GrayDecomp(Gray *gray);
void ElevateCoeffs(Gray *before, Gray *after, float scale, int minscale);
void GrayZeroAll(Gray *gray);
void ClearCoefs(int nbins);
void DumpCoefs(int nbins);
int CompareCoefs(Coeff *c1, Coeff *c2);
void CopyCoefs(int *bins, int nbins);
void RememberCoef(int coef, float val, int nbins);

#define MIN(a,b) (((a)<(b)) ? (a) : (b))

void GrayInit(void)			/* called automatically */
{
  int i;
  int lev = 0;

  levelTable[0] = 0;
  for (i = 1; i < MAXD; i++)
    {
      if (i >= sizes[lev]) 
	lev++;
      levelTable[i] = lev-1;
    }
}


void BuildScaleTable(float scale, int minscale)
{
  float entry;
  int   i,j,exp;

  for (i = 0; i < MAXL; i++)
    for (j = 0; j < MAXL; j++)
      {
	exp   = (minscale) ? MIN(i,j) : i+j;
	entry = pow(scale, (double) exp);
	scaleTable[i][j] = entry;
      }
}


void FailExit(char *msg)
{
  if (msg)
    fprintf(stderr, "fatal error: %s\n", msg);
  else
    fprintf(stderr, "fatal error!\n");

  exit(-1);
}


int SizeToLevel(int size)
{
  int i;
  
  for (i = 0; i < MAXL; i++)
    {
      if (size == sizes[i]) return i;
    }
  return BADL;			/* not found */
}


int LevelToSize(int level)
{
  return (1 << level);
}


void GrayFromImage(Gray *gray, Image *image)
{
  static int initialized = 0;
  int i,j;
  int size  = image->width;
  int level = SizeToLevel(size);

  if (!initialized)
    {
      GrayInit();
      initialized = 1;
    }

  if (image->width != image->height) 
    FailExit("image is not square");

  if (level == BADL) 
    FailExit("image size is not a power of two");

  if (!(image->type & PGM))
    FailExit("image is not pgm");

  gray->size  = size;
  gray->level = level;

  for (i = 0; i < size; i++)
    for (j = 0; j < size; j++)
      {
	gray->data[i][j] = ((float) GetPixel(image,j,i)) / ISCALE;
      }
}


void GrayToImage(Gray *gray, Image *image)
{
  float f, min, max;
  int i,j,c;
  int size = gray->size;
  int verbose = 0;

  if (image->width != image->height || image->width != size) 
    FailExit("image is not same size as gray");

  min = 1.0e20; max = -1.0e20;
  for (i=0; i < size; i++)
    { 
      for (j=0; j < size; j++)
        { 
          f = gray->data[i][j];
          if (f < min) min = f;
          if (f > max) max = f;
        }
    }

  for (i = 0; i < size; i++)
    {
      for (j = 0; j < size; j++)
	{
	  f = (gray->data[i][j] - min)/(max - min) * OSCALE;
	  c = (int)floor(f + 0.5);
	  if (verbose && i < 70 && j < 70)
	    {
	      if (f == 0.0)
		printf("X");
	      else
		printf("-");
	    }
	  SetPixel(image,j,i,c);
	}
      if (verbose && i < 70) printf("\n");
    }
}


#define AVERAGE(x,y) (((x)+(y))/2.0)


void CopyRow(float *before, float *after)
{
  int i;

  for (i = 0; i < MAXD; i++)
    {
      after[i] = before[i];
    }
}


void DecompSubRow(float *before, float *after, int size)
{
  float ave, diff;
  int i;
  int half = size / 2;

  if (DecompMode == MODE_SPACIAL)
    {
      for (i = 0; i < half; i++)
	{
	  ave           = AVERAGE(before[i*2], before[i*2+1]);
	  diff          = before[i*2] - ave;
	  after[i]      = ave;
	  after[half+i] = diff;
	}
    }
  else				/* freqency */
    {
      for (i = 0; i < half; i++)
	{
	  ave           = AVERAGE(before[i], before[half+i]);
	  diff          = before[i] - ave;
	  after[i]      = ave;
	  after[half+i] = diff;
	}
    }
}


void ReconSubRow(float *before, float *after, int size)
{
  float ave, diff;
  int i;
  int half = size / 2;

  if (DecompMode == MODE_SPACIAL)
    {
      for (i = 0; i < half; i++)
	{
	  ave          = before[i];
	  diff         = before[half+i];
	  after[2*i]   = ave + diff;
	  after[2*i+1] = ave - diff;
	}
    }
  else				/* frequency */
    {
      for (i = 0; i < half; i++)
	{
	  ave           = before[i];
	  diff          = before[half+i];
	  after[i]      = ave + diff;
	  after[half+i] = ave - diff;
	}
    }
}


void DecompMixedSubRow(float *before, float *after, int size, int p)
{
  float ave, diff;
  int i;
  int pp   = p - 1;
  int half = size / 2;
  int ii1, ii2;

  for (i = 0; i < half; i++)
    {
      ii1           = ((i & ~pp) << 1) | (i & pp);
      ii2           = ii1 | p;
      ave           = AVERAGE(before[ii1], before[ii2]);
      diff          = before[ii1] - ave;
      after[i]      = ave;
      after[half+i] = diff;
    }
}


void ReconMixedSubRow(float *before, float *after, int size, int p)
{
  float ave, diff;
  int i;
  int pp   = p - 1;
  int half = size / 2;
  int ii1, ii2;

  for (i = 0; i < half; i++)
    {
      ii1        = ((i & ~pp) << 1) | (i & pp);
      ii2        = ii1 | p;
      ave        = before[i];
      diff       = before[i+half];
      after[ii1] = ave + diff;
      after[ii2] = ave - diff;
    }
}


void DumpRow(float *data)
{
  int i;

  for (i = 0; i < MAXD; i++)
    {
      printf("%f ", data[i]);
    }
  printf("\n");
}


void DecompRow(float *data, int toplevel)
{
  static float temp[MAXD];
  int i;

  for (i = toplevel; i > 0; i--)
    {
      CopyRow(data, temp);
      DecompSubRow(temp, data, LevelToSize(i));
    }
}


void DecompMixedRow(float *data, int toplevel, int roworcol)
{
  static float temp[MAXD];
  int i,b,p;

  if (roworcol == ROW)
    b = MODE_MIXED_X(DecompMode);
  else
    b = MODE_MIXED_Y(DecompMode);

  if (b > MAXL-1) b = MAXL-1;

  p = 1 << b;

  for (i = toplevel; i > 0; i--)
    {
      CopyRow(data, temp);
      DecompMixedSubRow(temp, data, LevelToSize(i), p);
      p /= 2;
      if (p == 0) p = 1;
    }
}


void ReconRow(float *data, int toplevel)
{
  static float temp[MAXD];
  int i;

  for (i = 0; i <= toplevel; i++)
    {
      CopyRow(data, temp);
      ReconSubRow(temp, data, LevelToSize(i));
    }
}


void ReconMixedRow(float *data, int toplevel, int roworcol)
{
  static float temp[MAXD];
  int i,b,p;

  if (roworcol == ROW)
    b = MODE_MIXED_X(DecompMode);
  else
    b = MODE_MIXED_Y(DecompMode);

  if (b > MAXL-1) b = MAXL-1;

  p = 1;

  for (i = 0; i <= toplevel; i++)
    {
      CopyRow(data, temp);
      ReconMixedSubRow(temp, data, LevelToSize(i), p);
      if (i >= toplevel-b) p *= 2;
    }
}


void DecompRows(Gray *gray)
{
  int i;
  int size  = MAXD;
  int hilev = MAXL - 1;

  for (i = 0; i < size; i++)
    {
      DecompRow(gray->data[i], hilev);
    }
}


void DecompMixedRows(Gray *gray, int roworcol)
{
  int i;
  int size  = MAXD;
  int hilev = MAXL - 1;

  for (i = 0; i < size; i++)
    {
      DecompMixedRow(gray->data[i], hilev, roworcol);
    }
}


void ReconRows(Gray *gray)
{
  int i;
  int size  = MAXD;
  int hilev = MAXL - 1;

  for (i = 0; i < size; i++)
    {
      ReconRow(gray->data[i], hilev);
    }
}


void ReconMixedRows(Gray *gray, int roworcol)
{
  int i;
  int size  = MAXD;
  int hilev = MAXL - 1;

  for (i = 0; i < size; i++)
    {
      ReconMixedRow(gray->data[i], hilev, roworcol);
    }
}


void Transpose(Gray *gray)
{
  float t;
  int i,j;
  int size = gray->size;

  for (i = 0; i < size; i++)
    for (j = i+1; j < size; j++)
      {
	t                = gray->data[i][j];
	gray->data[i][j] = gray->data[j][i];
	gray->data[j][i] = t;
      }
}


void GraySetDecompMode(int mode)
{
  DecompMode = mode;
}


void GrayDecomp(Gray *gray)
{
  if (DecompMode <= MODE_FREQUENCY)
    {
      DecompRows(gray);
      Transpose(gray);
      DecompRows(gray);
      Transpose(gray);
    }
  else
    {
      DecompMixedRows(gray, ROW);
      Transpose(gray);
      DecompMixedRows(gray, COL);
      Transpose(gray);
    }
}


/*
 * make all wavelet coefficients be in range [0-1]
 */
void ElevateCoeffs(Gray *before, Gray *after, float scale, int minscale)
{
  float fac, max, c, ave;
  int i,j,lx,ly;
  int size = before->size;

  max = 0.0;
  ave = before->data[0][0];	/* save it for later */
  
  before->data[0][0] = 0.0;	/* don't mess up "max" below */

  /* precompute scales for all possible levels in hierarchy */
  BuildScaleTable(scale, minscale);

  for (i = 0; i < size; i++)
    for (j = 0; j < size; j++)
      {
	lx  = levelTable[j];
	ly  = levelTable[i];
	fac = scaleTable[ly][lx];
	c   = fabs(before->data[i][j]) / fac;

	after->data[i][j] = c;
	
	if (max < c) max = c;
      }

  for (i = 0; i < size; i++)
    for (j = 0; j < size; j++)
      {
	after->data[i][j] /= max;
      }

  after->size  = size;
  after->level = before->level;
  after->data[0][0] = ave;
}


/* 
 * convert a wavelet decomposition to an image.
 *
 * scale says by what factor f to scale down the low level wavelet coeffs.
 * for each coefficient c, there is a level in the x and y direction (lx & ly)
 * case minscale = 0:
 *      c <-- c / (2 ^ (lx+ly))
 * case minscale = 1:
 *      c <-- c / (2 ^ min(lx,ly))
 */
 
void GrayDecompToImage(Gray *gray, Image *image, float scale, int minscale)
{
  static Gray temp;

  ElevateCoeffs(gray, &temp, scale, minscale);
  GrayToImage(&temp, image);
}


void GrayZeroAll(Gray *gray)
{
  bzero((char *) gray->data, MAXD * MAXD * sizeof(float));
}


void GrayCopy(Gray *src, Gray *dest)
{
  *dest = *src;
}

/*
 * Globals for the next few routines:  (ugly, but faster)
 */

static Coeff topc[MAXB];
static float mintopc;
static int   mintopi;


void ClearCoefs(int nbins)
{
  int i;

  mintopc = 0.0;
  mintopi = 0;

  if (nbins > MAXB) FailExit("can't return that many coeffs");

  /* empty all bins */
  for (i = 0; i < nbins; i++)
    {
      topc[i].val = 0.0;
    }
}


void DumpCoefs(int nbins)	/* for debugging */
{
  int i;

  for (i = 0; i < nbins; i++)
    {
      printf("%d\t%f\n",topc[i].index,topc[i].val);
    }
  printf("\n");
}


/* used by qsort below */
int CompareCoefs(Coeff *c1, Coeff *c2)
{
  float v1 = c1->val;
  float v2 = c2->val;

  if (v1 == v2) return 0;
  if (v1 < v2)  return 1;
  return -1;
}


/* copy out indices of top coefficients */
void CopyCoefs(int *bins, int nbins)
{
  int i;

  /* sort them in decreasing order first */
  qsort((char *) topc, nbins, sizeof(struct Coeff), (Intfunc) CompareCoefs);

  /* copy out results */
  for (i = 0; i < nbins; i++)
    {
      bins[i] = topc[i].index;
    }
}


void RememberCoef(int coef, float val, int nbins)
{
  int j;

  topc[mintopi].val   = val;
  topc[mintopi].index = coef;
  mintopc             = 999.0;

  for (j = 0; j < nbins; j++) /* find new minimum */
    {
      if (mintopc > topc[j].val) 
	{
	  mintopc = topc[j].val;
	  mintopi = j;
	}
    }
  if (mintopc == 999.0) FailExit("coudn't find new minimum");
}


#define ROWCOL2INDEX(i,j) (MAXD*(i)+(j))
#define INDEX2ROW(i)      ((i)/MAXD)
#define INDEX2COL(i)      ((i)%MAXD)


/*
 * given a decomposed grayscale,
 * returns (in arrays x and y) the indices of the largest "nbins" coeffs
 */
void GrayDecompStats(Gray *gray, float scale, int minscale, 
		     int *bins, int nbins)
{
  static Gray  temp;
  float  c;
  int    i,j,coef;
  int    size = gray->size;

  ElevateCoeffs(gray, &temp, scale, minscale);

  temp.data[0][0] = 0.0;	/* we don't care about overall ave */

  ClearCoefs(nbins);

  /* now search whole shebang for the maximal entries */
  for (i = 0; i < size; i++)
    for (j = 0; j < size; j++)
      {
	c = temp.data[i][j];
	if (c > mintopc)
	  {
	    coef = ROWCOL2INDEX(i,j);
	    RememberCoef(coef, c, nbins);
	  }
      }

  CopyCoefs(bins, nbins);

  GrayCopy(&temp, gray);
}


void GrayZeroAllButNamed(Gray *gray, int *keep, int nbins)
{
  static Gray temp;
  int i,x,y;

  GrayZeroAll(&temp);
  
  temp.size  = gray->size;
  temp.level = gray->level;

  temp.data[0][0] = gray->data[0][0];

  for (i = 0; i < nbins; i++)
    {
      x = INDEX2COL(keep[i]);
      y = INDEX2ROW(keep[i]);
      temp.data[y][x] = gray->data[y][x];
    }

  GrayCopy(&temp, gray);
}


void GrayZeroAllSmall(Gray *gray, Gray *reference, float minsize)
{
  int  count = 0;
  int   size = gray->size;
  float  ave = gray->data[0][0];
  int i;
  for (i = 0; i < size; i++)
    { int j;
      for (j = 0; j < size; j++)
        {	if (reference->data[i][j] < minsize)
            { gray->data[i][j] = 0.0;
              count++;
            }
        }
    }
  gray->data[0][0] = ave;
  printf("count of zeroed coeffs: %d leaving %d\n", count, size*size-count);
}


void GrayRecon(Gray *gray)
{
  if (DecompMode <= MODE_FREQUENCY)
    {
      ReconRows(gray);
      Transpose(gray);
      ReconRows(gray);
      Transpose(gray);
    }
  else
    {
      ReconMixedRows(gray, ROW);
      Transpose(gray);
      ReconMixedRows(gray, COL);
      Transpose(gray);
    }
}
