/* test_sve_charges --- test of {sve_minn.h} with Rutherford's atom potential.  */
/* Last edited on 2009-02-22 19:43:15 by stolfi */

#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <values.h>
#include <jsstring.h>

#include <sve_minn.h>

#include <bool.h>
#include <jsrandom.h>
#include <argparser.h>
#include <affirm.h>
#include <rmxn.h>
#include <rmxn_extra.h>
#include <rn.h>
#include <jsfile.h>
#include <js.h>
#include <vec.h>

/* GENERAL PARAMETERS */

typedef struct options_t
  { int nq;   /* Number of electrons. */
    int dim;  /* Free dimensions. */
  } options_t;

#define MAXNQ 100
  /* Max number of electrons. */

/* INTERNAL PROTOTYPES */

int main (int argc, char **argv);

options_t *get_options(int argc, char **argv);
  /* Parses the command-line options. */

double electric_potential(double C, int nq, double Q[], int dim, double x[]);
  /* The electric potential of a set of pointlike `electrons' with
    charges {Q[0..nq-1]} at positions {x[0..dim*nq-1]}, immersed in
    a uniform spherical `nuclear cloud' with center at the
    origin whose total charge, up to unit distance from the origin, is
    {C}.
    
    Assumes three-dimensional cloud and {1/r^2} force in all cases,
    but only the first {dim} coordinates (1, 2, or 3) of each electron
    are unknown; the other {3 - dim} coordinates are implicitly fixed
    at 0 to be zero. */

void find_electron_positions(int nq, int dim);
  /* Computes coordinates {x[0..dim*nq-1]} that minimize the value of
     {electric_potential(C, nq, Q, dim, x)}, where all point electrons
     have charge {Q[i] = -1} and {C = -SUM{Q[i] : i \in 0..nq-1}}. */

void write_electron_positions(char *fname, int nq, int dim, double x[]);
  /* Writes the electron coordinates {x[0..dim*nq-1]} to a file named
    "{fname}". Each line contains an electron index and three coordinates
    (the first {dim} taken from {x}, and the other {3-dim} set to
    zero). If {fname} is NULL, writes to {stderr} instead. */

void plot_potential(char *fname, double C, int nq, double Q[], int dim, double R);
  /* Writes a file with sampled values of the electric potential as a
    function of the variable electron coordinates, in a format suitable
    for the {splot} command of {gnuplot}. The number of variable
    coordinates {dim*nq} must be at most 2. */

/* IMPLEMENTATIONS */

int main (int argc, char **argv)
  { options_t *o = get_options(argc, argv);
    find_electron_positions(o->nq, o->dim);
    fclose(stderr);
    fclose(stdout);
    return (0);
  }
  
void find_electron_positions(int nq, int dim)
  { 
    /* Output file names: */
    char *prefix = NULL; /* Prefix for output file names. */
    asprintf(&prefix, "test-%03d-%1d", nq, dim);
    char *solfname = txtcat(prefix, "-sol-ot"); /* Name of final solution file. */
    char *potfname = txtcat(prefix, "-pot-ot"); /* Name of potential field file. */
    
    int i; 
    /* Local storage: */
    double Q[nq];     /* Signed charge of each electron. */
    double C;         /* Total signed charge in nuclear cloud. */
    int nc = dim*nq;  /* Size of coordinate vector. */

    auto double F(int n, double x[]); 
      /* The goal function for uptimization. */
      
    double F(int n, double x[])
      { assert(n == nc);
        double Fx = electric_potential(C, nq, Q, dim, x);
        return Fx;
      }
      
    double Xprev[nc]; /* Guess in previous call of {OK} function. */
    int nok = 0;      /* Counts iterations (actually, calls to {OK}). */
    
    auto int OK(int n, double x[], double Fx); 
      /* Acceptance criterion function. */
      
    int OK(int n, double x[], double Fx)
      { assert(n == nc);
        fprintf(stderr, "iteration %d\n", nok);
        if (nok > 0)
          { double d = rn_dist(n, Xprev, x);
            fprintf(stderr, "displacement = %16.10f\n", d);
          }
        fprintf(stderr, "electron positions =\n");
        write_electron_positions(NULL, nq, dim, x);
        /* Save guess in {Xprev} for next call: */
        rn_copy(nc, x, Xprev); 
        nok++;
        fprintf(stderr, "potential = %16.10f\n", Fx);
        fprintf(stderr, "\n");
        return 0;
      }
      
    double x[nc];     /* Initial guess and final solution. */

    /* Define the electron charges {Q[0..nq-1]} and the total cloud charge {C}: */
    C = 0;
    for (i = 0; i < nq; i++) { Q[i] = -1; C += (-Q[i]); }
    
    /* Dump the potential field: */
    double RPlot = 3.0;  /* Half-width of plot doman. */
    if (nc <= 2) { plot_potential(potfname, C, nq, Q, dim, RPlot); }
    
    /* Initial electron coordinates: */
    srand(4615);  srandom(4615);
    for (i = 0; i < nc; i++) { x[i] = 2*drandom() - 1; }
    
    /* Optimize iteratively: */
    int NP = (nc+1)*(nc+2)/2; /* Number of probe points per iteration. */
    double rMin = 0.000001;
    double rMax = 10.0;
    double rIni = 1.00;
    int NITER = 100;
    int maxEvals = (NP+1)*NITER;
    bool_t debug = FALSE;
    sve_minn_iterate(nc, &F, &OK, x, rIni, rMin, rMax, INFINITY, maxEvals, debug);
    
    /* Ealuate (and print) the final potential: */
    double Fx = F(nc, x);
    
    /* Print final electron positions and write file for plotting: */
    fprintf(stderr, "final coordinates:\n");
    write_electron_positions(NULL, nq, dim, x);
    fprintf(stderr, "final potential = %16.10f\n", Fx);
    write_electron_positions(solfname, nq, dim, x);
    
    free(potfname);
    free(solfname);
    free(prefix);
  }

void plot_potential(char *fname, double C, int nq, double Q[], int dim, double R)
  { FILE *wr = open_write(fname, TRUE);
    int N = 20; /* Number of plot steps on each side of origin. */
    int nc = dim*nq;
    double x[nc];
    demand(nc >= 1, "too few coordinates to plot");
    double N0 = (nc < 1 ? 0 : N);
    double N1 = (nc < 2 ? 0 : N);
    demand(nc <= 2, "too many coordinates to plot");
    int i0, i1;
    for (i1 = -N1; i1 <= N1; i1++)
      { if (nc >= 2) { x[1] = (i1*R)/N1; }
        for (i0 = -N0; i0 <= +N0; i0++)
          { if (nc >= 1) { x[0] = (i0*R)/N0; }
            double Fx = electric_potential(C, nq, Q, dim, x);
            if (nc >= 1) { fprintf(wr, " %12.8f", x[0]); }
            if (nc >= 2) { fprintf(wr, " %12.8f", x[1]); }
            fprintf(wr, " %+22.16e\n", Fx);
          }
        if (N1 > 0) { fprintf(wr, "\n"); }
      }
    fflush(wr);
    fclose(wr);
  }

void write_electron_positions(char *fname, int nq, int dim, double x[])
  { FILE *wr;
    if (fname == NULL) 
      { wr = stderr; }
    else
      { wr = open_write(fname, TRUE);
        fprintf(wr, "# fields: index x Y Z\n");
      }
    int i, k;
    for (i = 0; i < nq; i++)
      { fprintf(wr, "%5d", i);
        for (k = 0; k < 3; k++)
          { double Xik = (k < dim ? x[dim*i + k] : 0);
            fprintf(wr, " %12.8f", Xik);
          }
        fprintf(wr, "\n");
      }
    fflush(wr);
    if ((wr != stderr) && (wr != stdout)) { fclose(wr); }
  }

double electric_potential(double C, int nq, double Q[], int dim, double x[])
  { int i, j, k;
    double Pcloud = 0; /* Potential electrons-vs-cloud. */
    double Ppairs = 0; /* Potential electrons-vs-electrons. */
    /* Fudge factors to avoid infinities and nans: */
    double relfudge = 1.0e-6; /* Relative fudge factor. */
    double absfudge = 1.0e-6; /* Absolute fudge term. */
    for (i = 0; i < nq; i++)
      { /* Add the potential due to the cloud on electron {i}: */
        double r2 = 0;
        for (k = 0; k < 3; k++) 
          { double Xik = (k < dim ? x[dim*i + k] : 0); 
            r2 += Xik*Xik;
          }
        double r = sqrt(r2);
        /* Compute the total charge {Ci} in the cloud up to radius {r}: */
        double Ci = C*(4*M_PI/3)*r2*r;
        double Pi = (r == 0.0 ? 0.0 : -Ci*Q[i]/r);
        /* fprintf(stderr, "  ri =  %8.4f Pi =  %8.4f\n", r, Pi); */
        Pcloud += Pi;
        /* Add the potential of electrons {0..i-1} on electron {i}: */
        for (j = 0; j < i; j++) 
          { double dij2 = 0; /* Squared distance between electrons {i} and {j}. */
            double rij2 = 0; /* Squared tot distance between electrons {i,j} and origin. */
            for (k = 0; k < 3; k++) 
              { double Xik = (k < dim ? x[dim*i+k] : 0); 
                double Xjk = (k < dim ? x[dim*j+k] : 0); 
                double dijk = Xik - Xjk;
                dij2 += dijk*dijk;
                rij2 += Xik*Xik + Xjk+Xjk;
              }
            double dij = sqrt(dij2 + 2*relfudge*rij2 + absfudge*absfudge);
            double Pij = Q[i]*Q[j]/dij;
            /* fprintf(stderr, "  dij = %8.4f Pij = %8.4f\n", d, Pij); */
            Ppairs += Pij;
          }
      }
    /* fprintf(stderr, "  Pcloud =  %8.4f Ppairs =  %8.4f\n", Pcloud, Ppairs); */
    return Pcloud + Ppairs;
  }
  
options_t *get_options(int argc, char **argv)
  { argparser_t *pp = argparser_new(stderr, argc, argv);
    options_t *o = notnull(malloc(sizeof(options_t)), "no mem"); 
    o->nq = argparser_get_next_int(pp, 1, MAXNQ);
    o->dim = argparser_get_next_int(pp, 1, 3);
    argparser_finish(pp);
    return o;
  }
