#define PROG_NAME "test_sve_img_match"
#define PROG_DESC "tests {sve_minn.h} on an image matching problem"
#define PROG_VERS "1.0"

/* Last edited on 2009-02-22 19:43:24 by stolfi */

#define test_sve_img_match_C_COPYRIGHT "Copyright  2009 by the State University of Campinas (UNICAMP)"

#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <values.h>

#include <bool.h>
#include <float_image.h>
#include <float_image_transform.h>
#include <argparser.h>
#include <r3x3.h>
#include <r3.h>
#include <rn.h>
#include <affirm.h>
#include <jsfile.h>
#include <js.h>
#include <vec.h>

#define PROG_HELP = \
  "  " PROG_NAME " \\\n" \
  "    -img {IMG_NAME}.pgm \\\n" \
  "    -obj {OBJ_NAME}.pgm \\\n" \
  "    [ -msk {MSK_NAME}.pgm ]"

#define PROG_INFO \
  "NAME\n" \
  "  " PROG_NAME " - " PROG_DESC "\n" \
  "\n" \
  "SYNOPSIS\n" \
  PROG_HELP "\n" \
  "\n" \
  "DESCRIPTION\n" \
  "  Tests {sve_minn.h} on an image matching problem.  The" \
  " parameter is a 3x3 projective transformation matrix {M}.  The" \
  " goal function is the quadratic mismatch between a given" \
  " image {obj, windowed by an optional mask {msk} and" \
  " transformed by {M}, and a second image {img}.\n" \
  "\n" \
  "  The three images {img}, {obj}, and {msk} are read" \
  " from the files \"{IMG_NAME}.png\", \"{OBJ_NAME}.png\"," \
  " and \"{MSK_NAME}.png\", respectively.  If the mask" \
  " is not specified, uses a 2D Hann window.\n" \
  "\n" \
  "OPTIONS\n" \
  "  See above.\n" \
  "\n" \
  "DOCUMENTATION OPTIONS\n" \
  argparser_help_info_HELP_INFO "\n" \
  "\n" \
  "SEE ALSO\n" \
  "  Hamlet(10), Venice(11), Luxor(11), Amazon(12).\n" \
  "\n" \
  "AUTHOR\n" \
  "  Created 2009-01-05 by Jorge Stolfi, IC-UNICAMP.\n" \
  "\n" \
  "MODIFICATION HISTORY\n" \
  "\n" \
  "WARRANTY\n" \
  argparser_help_info_NO_WARRANTY "\n" \
  "\n" \
  "RIGHTS\n" \
  "  " test_sve_img_match_C_COPYRIGHT ".\n" \
  "\n" \
  argparser_help_info_STANDARD_RIGHTS


#include <sve_minn.h>

/* GENERAL PARAMETERS */

typedef struct options_t
  { char *img_name;  /* Target image. */
    char *obj_name;  /* Object image. */
    char *msk_name;  /* Mask image. */
  } options_t;

#define MAXNQ 100
  /* Max number of . */

/* INTERNAL PROTOTYPES */

int main (int argc, char **argv);

options_t *get_options(int argc, char **argv);
  /* Parses the command-line options. */

float_image_t *read_input_image(char *img_name, int *NCP, int *NXP, int *NYP);
  /* If {img_name} is NULL, returns NULL. 
  
    Otherwise reads a grayscale image from file "{img_name}", which
    must be in PGM format. Then, if {*NCP} is negative, sets {*NCP} to
    the number of channels {NC} in the image; otherwise requires {*NCP
    == NC}. Ditto for {*NXP} and {NYP}. */

double mismatch
  ( float_image_t *img, 
    float_image_t *obj, 
    float_image_t *msk, 
    r3x3_t *M
  );
  /* Relative mismatch between the image {obj}, weighted by the mask {msk}
    and transformed by {M}, and the image {img}.
    
    The mismatch is defined as {(1/A)\int\int msk(p)*(obj(p) -
    img(M(p)))^2 dp}, where the integration is over the domain of
    {obj} (which must be the domain of {msk}), and {A} is the area of
    that domain. */

void param_to_matrix(int n, double x[], r3x3_t *M);
  /* Fills the matrix {M} with a "1" in position {[0][0]}, and with
    {x[0..7]} in the remaining positions, row by row. */

void matrix_to_param(r3x3_t *M, int n, double x[]);
  /* Ignores element {[0][0]} of matrix {M}, and collects the
    remaining elements in {x[0..7]}, row by row. */

void find_best_matrix
  ( float_image_t *img, 
    float_image_t *obj, 
    float_image_t *msk, 
    r3x3_t *M
  );
  /* Finds the matrix {M} that minimizes {mismatch(img,obj,msk)}.
    On input, {M} must be the initial guess.  On output, 
    {M} will be the optimal matrix as found by the method. */

void write_matched_object_image
  ( char *out_name,
    float_image_t *img, 
    float_image_t *obj, 
    float_image_t *msk, 
    r3x3_t *M,
    float bgv[],
    bool_t object
  );
  /* Writes an image that illustrates the matching of image {obj},
    masked by {msk} and transformed by {M}, against image {img}.
    
    The output image consists of a background image {bgr} blended with
    a second image {val}, with weights {1-msk'} and {msk'}, where
    {msk'} is {msk} transformed by {M}.
    
    The background {bgr} has the same size as {img} and is filled with
    the pixel value {v[0..nc-1]}. If {object} is false, the second
    image {val} is the frame image {img}. If {object} is true, the
    image {val} is {obj} transformed by {M}. 
    
    The image is written to file "{out_name}.pgm"*/

/* IMPLEMENTATIONS */

int main (int argc, char **argv)
  { options_t *o = get_options(argc, argv);
  
    /* Read the target image {img}: */
    int NC = -1; /* Number of channels in {img,obj}. */
    int NXI = -1, NYI = -1; /* Size of {img}. */
    float_image_t *img = read_input_image(o->img_name, &NC, &NXI, &NYI);
    
    /* Read the object image {obj}: */
    int NXO = -1, NYO = -1; /* Size of {obj} and {msk}. */
    float_image_t *obj = read_input_image(o->obj_name, &NC, &NXO, &NYO);

    /* Read or manufacture the mask image {msk}: */
    int NCM = 1;  /* Number of channels in {msk}. */
    float_image_t *msk = read_input_image(o->msk_name, &NCM, &NXO, &NYO);
    if (msk == NULL) { msk = ; }
    float_image_masked_t mob = (float_image_masked_t) { .img = obj, .msk = msk };
    
    /* Find and print the optimum transform: */
    r3x3_t M;
    r3x3_ident(&M);
    find_best_matrix(img, obj, msk, &M);
    
    /* Write the diagnostic images: */
    float bgv[NC]; /* Background pixel value for outout images. */
    int c; for (c = 0; c < NC; c++) { bgv[c] = 0.5; }
    write_matched_object_image("out/msk-obj", img, obj, msk, &M, bgv, TRUE);
    write_matched_object_image("out/msk-img", img, obj, msk, &M, bgv, FALSE);
    
    fclose(stdout);
    return (0);
  }
  
  float_image_masked_interpolate
  ( float_image_masked_t *im, 
    int c,
    double x,
    double y,
    int degInter,
    float *val,
    float *wht
  )



   float_image_paint_sample
  ( float_image_t *A, 
    int c, 
    int ix, 
    int iy, 
    float_image_func_t *func, 
    float_image_func_t *mask, 
    int m
  );

float_image_t *read_input_image(char *img_name, int *NCP, int *NXP, int *NYP)
  {
    if (img_name == NULL)
      { return NULL; }
    else
      { FILE *rd = open_read(img_name, TRUE);
        float_image_t *img = float_image_read(rd);
        fclose(rd);
        return img;
      }
  }

void param_to_matrix(int n, double x[], r3x3_t *M)
  { assert(n == 8);
    int k = 0;
    int i, j;
    for (i = 0; i < 3; i++)
      { for (j = 0; j < 3; j++)
          { if ((i == 0) && (j == 0))
              { M->c[i][j] = 1.0; }
            else
              { M->c[i][j] = x[k]; k++; }
          }
      }
  }

void matrix_to_param(r3x3_t *M, int n, double x[])
  { assert(n == 8);
    int k = 0;
    int i, j;
    for (i = 0; i < 3; i++)
      { for (j = 0; j < 3; j++)
          { if ((i == 0) && (j == 0))
              { /* Skip element. */ }
            else
              { x[k] = M->c[i][j]; k++; }
          }
      }
  }

void find_best_matrix
  ( float_image_t *img, 
    float_image_t *obj, 
    float_image_t *msk, 
    r3x3_t *M
  )
  { 
    int ne = 8; /* Number of variable elements in matrix. */
    
    auto double F(int n, double x[]); 
      /* The goal function for uptimization. */
      
    double F(int n, double x[])
      { assert(n == ne);
        r3x3_t M;
        param_to_matrix(n, x, &M);
        double Fx = mismatch(img, obj, msk, &M);
        return Fx;
      }
      
    double xPrev[ne]; /* Guess in previous call of {OK} function. */
    int nok = 0;      /* Counts iterations (actually, calls to {OK}). */
    
    auto int OK(int n, double x[], double Fx); 
      /* Acceptance criterion function. */
      
    int OK(int n, double x[], double Fx)
      { assert(n == ne);
        fprintf(stderr, "iteration %d\n", nok);
        if (nok > 0)
          { double d = rn_dist(n, xPrev, x);
            fprintf(stderr, "change = %16.10f\n", d);
          }
        r3x3_t M;
        param_to_matrix(n, x, &M);
        fprintf(stderr, "matrix =\n");
        r3x3_print(stderr, &M);
        /* Save guess in {xPrev} for next call: */
        rn_copy(ne, x, xPrev); 
        nok++;
        fprintf(stderr, "mismatch = %16.10f\n", Fx);
        fprintf(stderr, "\n");
        return 0;
      }
      
    double x[ne];     /* Initial guess and final solution. */

    /* Start with the identity matrix: */
    r3x3_t T;
    r3x3_ident(&T);
    matrix_to_param(&T, ne, x);
    
    /* Optimize iteratively: */
    int NP = (ne+1)*(ne+2)/2; /* Number of probe points per iteration. */
    double rMin = 0.000001;
    double rMax = 10.0;
    double rIni = 0.20;
    int NITER = 100;
    int maxEvals = (NP+1)*NITER;
    sve_minn_iterate(ne, &F, &OK, x, rIni, rMin, rMax, INFINITY, maxEvals);
    
    /* Ealuate (and print) the final mismatch: */
    double Fx = F(ne, x);
    
    /* Print final matrix: */
    param_to_matrix(ne, x, &T);
    fprintf(stderr, "final matrix =\n");
    r3x3_print(stderr, &T);
    fprintf(stderr, "final potential = %16.10f\n", Fx);
  }
 
void write_matched_object_image
  ( char *out_name,
    float_image_t *img, 
    float_image_t *obj, 
    float_image_t *msk, 
    r3x3_t *M,
    float bgv[],
    bool_t object
  )
  {
    /* Get the input image dimensions: */
    int NC = img->sz[0];            /* Number of channels in all images. */
    assert(obj->sz[0] == NC);
    if (msk != NULL) { assert(msk->sz[0] == NC); }
    
    int NXI = img->sz[1], NYI = img->sz[2]; /* Size of {img}. */

    int NXO = obj->sz[1], NYO = obj->sz[2]; /* Size of {obj} and {msk}. */
    if (msk != NULL) { assert(msk->sz[1] == NXO); assert(msk->sz[2] == NYO); }
    
    /* Create the output image an fill it with the background pixel value: */
    float_image_t *out = float_image_new (NC, NXI, NYI);
    float_image_fill_pixels(out, c, bgv);
    
    /* Paste the image {obj} onto {
  }

options_t *get_options(int argc, char **argv)
  { argparser_t *pp = argparser_new(stderr, argc, argv);
    options_t *o = notnull(malloc(sizeof(options_t)), "no mem");
    
    argparser_get_keyword(pp, "-img");
    o->img_name = argparser_get_next(pp);
    
    argparser_get_keyword(pp, "-obj");
    o->obj_name = argparser_get_next(pp);
    
    if (argparser_keyword_present(pp, "-msk"))
      { o->msk_name = NULL; }
    else
      { o->msk_name = argparser_get_next(pp); }
    
    argparser_finish(pp);
    return o;
  }
