/* See udg_grid.h */
/* Last edited on 2009-08-22 18:21:02 by stolfi */

#define _GNU_SOURCE
#include <affirm.h>
#include <bool.h>
#include <indexing.h>
#include <stdlib.h>
#include <assert.h>
  
#include <udg_grid.h>

udg_grid_size_t udg_grid_size(udg_rank_t r)
  { return (1LL << r);  }

udg_grid_pos_t udg_max_grid_pos(udg_rank_t r)
  { return (1LL << r) - 1; }

sign_t udg_rank_cmp(udg_rank_t *x, udg_rank_t *y)
  { if ((*x) < (*y))
      { return -1; }
    else if ((*x) > (*y))
      { return +1; }
    else
      { return 0; }
  }
  
/* CELL INDICES */
  
udg_rank_t udg_cell_rank(udg_cell_index_t k)
  { int r = 0;
    while (k > (1LL << 16)) { k >>= 16; r+= 16; }
    while (k > (1LL << 4)) { k >>= 4; r+= 4; }
    while (k > 1) { k >>= 1; r+= 1; }
    return r;
  }

udg_grid_pos_t udg_cell_position(udg_cell_index_t k)
  { udg_cell_index_t t = ((1LL << udg_cell_rank(k)) >> 1);
    return k - t;
  }
  
udg_cell_index_t udg_cell_from_position(udg_rank_t r, udg_grid_pos_t pos)
  { /* First cell of layer {r}: */
    udg_cell_index_t k0 = (1LL << r);
    /* Number of cells {sz} in layer {r}: */
    udg_cell_index_t sz = (1LL << r);
    /* Displacement {t} from {k0} is {pos % sz}: */
    uint64_t t = pos & (sz - 1);
    /* Compute index: */
    return k0 + t;
  }

udg_cell_index_t udg_cell_shift(udg_cell_index_t k, udg_grid_pos_t dp)
  { udg_rank_t r = udg_cell_rank(k);
    /* First cell of layer {r}: */
    udg_cell_index_t k0 = (1LL << r);
    /* Number of cells {sz} in layer {r}: */
    udg_cell_index_t sz = (1LL << r);
    /* Displacement {t} from {k0} is {(k - k0 + dp) % sz}: */
    uint64_t t = (k - k0 + dp) & (sz - 1);
    /* Compute index: */
    return k0 + t;
  }

void udg_cell_interval_root_relative(udg_cell_index_t k, interval_t *B)
  { udg_grid_pos_t pos = udg_cell_position(k);
    udg_grid_size_t gsz = udg_grid_size(udg_cell_rank(k));
    double scale = 1.0/((double)gsz);
    LO((*B)) = scale*((double)pos);
    HI((*B)) = scale*((double)(pos + 1));
  }

/* CELL PACKS */

void udg_pack_cells(udg_cell_index_t b, udg_grid_size_t psz, udg_cell_index_t k[])
  { udg_pack_slot_index_t ix = 0;
    for (ix = 0; ix < psz; ix++)
      { k[ix] = udg_cell_shift(b, ix); }
  }

udg_pack_slot_index_t udg_pack_find_cell
  ( udg_cell_index_t k, 
    udg_cell_index_t b,
    udg_grid_size_t psz
  )
  {
    /* Get ranks of {k} and {b}: */
    udg_rank_t kr = udg_cell_rank(k);
    udg_rank_t br = udg_cell_rank(b);
    if (kr != br) { return udg_NOT_SLOT; }

    /* Get position of {k} and {b} in that layer: */
    udg_grid_pos_t kp = udg_cell_position(k);
    udg_grid_pos_t bp = udg_cell_position(b);
 
    /* Get position of {k} relative to {b}, beware of wrap-around: */
    udg_grid_size_t gsz = udg_grid_size(br);
    udg_grid_pos_t e = (kp - bp + gsz) % gsz;
    if (e >= psz) { return udg_NOT_SLOT; }
    return (udg_pack_slot_index_t)e;
  }

bool_t udg_pack_intersects_cell
  ( udg_cell_index_t k, 
    udg_cell_index_t b,
    udg_grid_size_t psz
  )
  {
    /* Get rank of {b}, its position in that layer, and size of layer: */
    udg_rank_t br = udg_cell_rank(b);
    udg_grid_pos_t bp = udg_cell_position(b);
    udg_grid_size_t bgsz = udg_grid_size(br);
    
    /* Get rank of {k}: */
    udg_rank_t kr = udg_cell_rank(k);
    
    /* If {k} is deeper than {b}, get its ancestor of same rank: */
    if (kr > br) { k = k >> (kr - br); kr = br; }
    
    /* Get positions of min and max descendants of {k} on level {Er}: */
    udg_grid_pos_t kp_lo = udg_cell_position(k << (br - kr));
    udg_grid_pos_t kp_hi = udg_cell_position(((k + 1) << (br - kr)) - 1);
    assert(kp_lo <= kp_hi); /* A single cell can't wrap around. */

    /* Get rel position range {pp_lo..pp_hi} of pack in level {br}: */ 
    udg_grid_pos_t pp_lo = bp;
    udg_grid_pos_t pp_hi = (pp_lo + psz - 1) % bgsz;
    
    /* Make sure {pp_hi >= pp_lo}: */
    if (pp_hi < pp_lo) { pp_hi += bgsz; }
    
    /* Check whether the ranges intersect (with or without wrap-around): */
    if (kp_lo < pp_lo)
      { if ((kp_hi < pp_lo) && (pp_hi < kp_lo + bgsz)) { return FALSE; } }
    else if (pp_lo < kp_lo)
      { if (kp_lo > pp_hi) { return FALSE; }}
    
    return TRUE;
  }

void udg_pack_interval_root_relative
  ( udg_cell_index_t b,
    udg_grid_size_t psz, 
    interval_t *B
  )
  { udg_cell_interval_root_relative(b, B);
    HI((*B)) = HI((*B)) + (psz-1)*(HI((*B)) - LO((*B)));
  }

/* PRINTOUT */

void udg_cell_print(FILE *wr, udg_cell_index_t k)
  { 
    udg_grid_pos_t kp = udg_cell_position(k);
    udg_rank_t kr = udg_cell_rank(k);
    fprintf(wr, "%d:", (int)kr);
    fprintf(wr, "%d", (int)kp);
  }
  
/* MANIPULATION OF CELL INDEX VECTORS */

vec_typeimpl(udg_cell_index_vec_t,udg_cell_index_vec,udg_cell_index_t);
